"""BIDS compatible path functionality."""

# Authors: Adam Li <adam2392@gmail.com>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#
# License: BSD-3-Clause
import glob
import json
import os
import re
import shutil as sh
from copy import deepcopy
from datetime import datetime
from io import StringIO
from os import path as op
from pathlib import Path
from textwrap import indent
from typing import Optional

import numpy as np
from mne.utils import _check_fname, _validate_type, logger, verbose

from mne_bids.config import (
    ALLOWED_DATATYPE_EXTENSIONS,
    ALLOWED_DATATYPES,
    ALLOWED_FILENAME_EXTENSIONS,
    ALLOWED_FILENAME_SUFFIX,
    ALLOWED_PATH_ENTITIES,
    ALLOWED_PATH_ENTITIES_SHORT,
    ALLOWED_SPACES,
    ENTITY_VALUE_TYPE,
    reader,
)
from mne_bids.tsv_handler import _drop, _from_tsv, _to_tsv
from mne_bids.utils import (
    _check_empty_room_basename,
    _check_key_val,
    _ensure_tuple,
    param_regex,
    warn,
)


def _find_empty_room_candidates(bids_path):
    """Get matching empty-room file for an MEG recording."""
    # Check whether we have a BIDS root.
    bids_root = bids_path.root
    if bids_root is None:
        raise ValueError(
            'The root of the "bids_path" must be set. '
            'Please use `bids_path.update(root="<root>")` '
            "to set the root of the BIDS folder to read."
        )

    bids_path = bids_path.copy()

    datatype = "meg"  # We're only concerned about MEG data here
    bids_fname = bids_path.update(suffix=datatype).fpath
    _, ext = _parse_ext(bids_fname)
    emptyroom_dir = BIDSPath(root=bids_root, subject="emptyroom").directory

    if not emptyroom_dir.exists():
        return list()

    # Find the empty-room recording sessions.
    emptyroom_session_dirs = [
        x
        for x in emptyroom_dir.iterdir()
        if x.is_dir() and str(x.name).startswith("ses-")
    ]
    if not emptyroom_session_dirs:  # No session sub-directories found
        emptyroom_session_dirs = [emptyroom_dir]

    # Now try to discover all recordings inside the session directories.

    allowed_extensions = list(reader.keys())
    # `.pdf` is just a "virtual" extension for BTi data (which is stored inside
    # a dedicated directory that doesn't have an extension)
    del allowed_extensions[allowed_extensions.index(".pdf")]

    candidate_er_fnames = []
    for session_dir in emptyroom_session_dirs:
        dir_contents = glob.glob(
            op.join(session_dir, datatype, f"sub-emptyroom_*_{datatype}*")
        )
        for item in dir_contents:
            item = Path(item)
            if (item.suffix in allowed_extensions) or (
                not item.suffix and item.is_dir()
            ):  # Hopefully BTi?
                candidate_er_fnames.append(item.name)

    candidates = list()
    for er_fname in candidate_er_fnames:
        # get entities from filenamme
        er_bids_path = get_bids_path_from_fname(er_fname, check=False)
        er_bids_path.subject = "emptyroom"  # er subject entity is different
        er_bids_path.root = bids_root
        er_bids_path.datatype = "meg"
        candidates.append(er_bids_path)

    return candidates


def _find_matched_empty_room(bids_path):
    from mne_bids import read_raw_bids  # avoid circular import.

    candidates = _find_empty_room_candidates(bids_path)

    # Walk through recordings, trying to extract the recording date:
    # First, from the filename; and if that fails, from `info['meas_date']`.
    best_er_bids_path = None
    min_delta_t = np.inf
    date_tie = False
    failed_to_get_er_date_count = 0
    bids_path = bids_path.copy().update(datatype="meg")
    raw = read_raw_bids(bids_path=bids_path)
    if raw.info["meas_date"] is None:
        raise ValueError(
            "The provided recording does not have a measurement "
            "date set. Cannot get matching empty-room file."
        )
    ref_date = raw.info["meas_date"]
    del bids_path, raw
    for er_bids_path in candidates:
        # get entities from filenamme
        er_meas_date = None

        # Try to extract date from filename.
        if er_bids_path.session is not None:
            try:
                er_meas_date = datetime.strptime(er_bids_path.session, "%Y%m%d")
            except (ValueError, TypeError):
                # There is a session in the filename, but it doesn't encode a
                # valid date.
                pass

        if er_meas_date is None:  # No luck so far! Check info['meas_date']
            _, ext = _parse_ext(er_bids_path.fpath)
            extra_params = None
            if ext == ".fif":
                extra_params = dict(allow_maxshield="yes")

            er_raw = read_raw_bids(bids_path=er_bids_path, extra_params=extra_params)

            er_meas_date = er_raw.info["meas_date"]
            if er_meas_date is None:  # There's nothing we can do.
                failed_to_get_er_date_count += 1
                continue

        er_meas_date = er_meas_date.replace(tzinfo=ref_date.tzinfo)
        delta_t = er_meas_date - ref_date

        if abs(delta_t.total_seconds()) == min_delta_t:
            date_tie = True
        elif abs(delta_t.total_seconds()) < min_delta_t:
            min_delta_t = abs(delta_t.total_seconds())
            best_er_bids_path = er_bids_path
            date_tie = False

    if failed_to_get_er_date_count > 0:
        msg = (
            f"Could not retrieve the empty-room measurement date from "
            f"a total of {failed_to_get_er_date_count} recording(s)."
        )
        warn(msg)

    if date_tie:
        msg = (
            "Found more than one matching empty-room measurement with the "
            "same recording date. Selecting the first match."
        )
        warn(msg)

    return best_er_bids_path


class BIDSPath:
    """A BIDS path object.

    BIDS filename prefixes have one or more pieces of metadata in them. They
    must follow a particular order, which is followed by this function. This
    will generate the *prefix* for a BIDS filename that can be used with many
    subsequent files, or you may also give a suffix that will then complete
    the file name.

    BIDSPath allows dynamic updating of its entities in place, and operates
    similar to `pathlib.Path`. In addition, it can query multiple paths
    with matching BIDS entities via the ``match`` method.

    Note that not all parameters are applicable to each suffix of data. For
    example, electrode location TSV files do not need a "task" field.

    Parameters
    ----------
    subject : str | None
        The subject ID. Corresponds to "sub".
    session : str | None
        The acquisition session. Corresponds to "ses".
    task : str | None
        The experimental task. Corresponds to "task".
    acquisition: str | None
        The acquisition parameters. Corresponds to "acq".
    run : int | None
        The run number. Corresponds to "run".
    processing : str | None
        The processing label. Corresponds to "proc".
    recording : str | None
        The recording name. Corresponds to "rec".
    space : str | None
        The coordinate space for anatomical and sensor location
        files (e.g., ``*_electrodes.tsv``, ``*_markers.mrk``).
        Corresponds to "space".
        Note that valid values for ``space`` must come from a list
        of BIDS keywords as described in the BIDS specification.
    split : int | None
        The split of the continuous recording file for ``.fif`` data.
        Corresponds to "split".
    description : str | None
        This corresponds to the BIDS entity ``desc``. It is used to provide
        additional information for derivative data, e.g., preprocessed data
        may be assigned ``description='cleaned'``.

        .. versionadded:: 0.11
    suffix : str | None
        The filename suffix. This is the entity after the
        last ``_`` before the extension. E.g., ``'channels'``.
        The following filename suffix's are accepted:
        'meg', 'markers', 'eeg', 'ieeg', 'T1w',
        'participants', 'scans', 'electrodes', 'coordsystem',
        'channels', 'events', 'headshape', 'digitizer',
        'beh', 'physio', 'stim'
    extension : str | None
        The extension of the filename. E.g., ``'.json'``.
    datatype : str
        The BIDS data type, e.g., ``'anat'``, ``'func'``, ``'eeg'``, ``'meg'``,
        ``'ieeg'``.
    root : path-like | None
        The root directory of the BIDS dataset.
    check : bool
        If ``True``, enforces BIDS conformity. Defaults to ``True``.

    Attributes
    ----------
    entities : dict
        A dictionary of the BIDS entities and their values:
        ``subject``, ``session``, ``task``, ``acquisition``,
        ``run``, ``processing``, ``space``, ``recording``,
        ``split``, ``description``, ``suffix``, and ``extension``.
    datatype : str | None
        The data type, i.e., one of ``'meg'``, ``'eeg'``, ``'ieeg'``,
        ``'anat'``.
    basename : str
        The basename of the file path. Similar to `os.path.basename(fpath)`.
    root : pathlib.Path
        The root of the BIDS path.
    directory : pathlib.Path
        The directory path.
    fpath : pathlib.Path
        The full file path.
    check : bool
        Whether to enforce BIDS conformity.

    Examples
    --------
    Generate a BIDSPath object and inspect it

    >>> bids_path = BIDSPath(subject='test', session='two', task='mytask',
    ...                      suffix='ieeg', extension='.edf', datatype='ieeg')
    >>> print(bids_path.basename)
    sub-test_ses-two_task-mytask_ieeg.edf
    >>> bids_path
    BIDSPath(
    root: None
    datatype: ieeg
    basename: sub-test_ses-two_task-mytask_ieeg.edf)

    Copy and update multiple entities at once

    >>> new_bids_path = bids_path.copy().update(subject='test2',
    ...                                         session='one')
    >>> print(new_bids_path.basename)
    sub-test2_ses-one_task-mytask_ieeg.edf

    Printing a BIDSPath will show a relative path when `root` is not set

    >>> print(new_bids_path)
    sub-test2/ses-one/ieeg/sub-test2_ses-one_task-mytask_ieeg.edf

    Setting `suffix` without an identifiable datatype will make
    BIDSPath try to guess the datatype

    >>> new_bids_path = new_bids_path.update(suffix='channels',
    ...                                      extension='.tsv')
    >>> print(new_bids_path)
    sub-test2/ses-one/ieeg/sub-test2_ses-one_task-mytask_channels.tsv

    You can set a new root for the BIDS dataset. Let's see what the
    different properties look like for our object:

    >>> new_bids_path = new_bids_path.update(root='/bids_dataset')
    >>> print(new_bids_path.root.as_posix())
    /bids_dataset
    >>> print(new_bids_path.basename)
    sub-test2_ses-one_task-mytask_channels.tsv
    >>> print(new_bids_path)
    /bids_dataset/sub-test2/ses-one/ieeg/sub-test2_ses-one_task-mytask_channels.tsv
    >>> print(new_bids_path.directory.as_posix())
    /bids_dataset/sub-test2/ses-one/ieeg

    Notes
    -----
    BIDS entities are generally separated with a ``"_"`` character, while
    entity key/value pairs are separated with a ``"-"`` character.
    There are checks performed to make sure that there are no ``'-'``, ``'_'``,
    or ``'/'`` characters contained in any entity keys or values.

    To represent a filename such as ``dataset_description.json``,
    one can set ``check=False``, and pass ``suffix='dataset_description'``
    and ``extension='.json'``.

    ``BIDSPath`` can also be used to represent file and folder names of data
    types that are not yet supported through MNE-BIDS, but are recognized by
    BIDS. For example, one can set ``datatype`` to ``dwi`` or ``func`` and
    pass ``check=False`` to represent diffusion-weighted imaging and
    functional MRI paths.
    """

    def __init__(
        self,
        subject=None,
        session=None,
        task=None,
        acquisition=None,
        run=None,
        processing=None,
        recording=None,
        space=None,
        split=None,
        description=None,
        root=None,
        suffix=None,
        extension=None,
        datatype=None,
        check=True,
    ):
        if all(
            ii is None
            for ii in [
                subject,
                session,
                task,
                acquisition,
                run,
                processing,
                recording,
                space,
                description,
                root,
                suffix,
                extension,
            ]
        ):
            raise ValueError("At least one parameter must be given.")

        self.check = check

        self.update(
            subject=subject,
            session=session,
            task=task,
            acquisition=acquisition,
            run=run,
            processing=processing,
            recording=recording,
            space=space,
            split=split,
            description=description,
            root=root,
            datatype=datatype,
            suffix=suffix,
            extension=extension,
        )

    @property
    def entities(self):
        """Return dictionary of the BIDS entities."""
        return {
            "subject": self.subject,
            "session": self.session,
            "task": self.task,
            "acquisition": self.acquisition,
            "run": self.run,
            "processing": self.processing,
            "space": self.space,
            "recording": self.recording,
            "split": self.split,
            "description": self.description,
        }

    @property
    def basename(self):
        """Path basename."""
        basename = []
        for key, val in self.entities.items():
            if val is not None and key != "datatype":
                # convert certain keys to shorthand
                long_to_short_entity = {
                    val: key for key, val in ALLOWED_PATH_ENTITIES_SHORT.items()
                }
                key = long_to_short_entity[key]
                basename.append(f"{key}-{val}")

        if self.suffix is not None:
            if self.extension is not None:
                basename.append(f"{self.suffix}{self.extension}")
            else:
                basename.append(self.suffix)

        basename = "_".join(basename)
        return basename

    @property
    def directory(self):
        """Get the BIDS parent directory.

        If ``subject``, ``session`` and ``datatype`` are set, then they will be
        used to construct the directory location. For example, if
        ``subject='01'``, ``session='02'`` and ``datatype='ieeg'``, then the
        directory would be::

            <root>/sub-01/ses-02/ieeg

        Returns
        -------
        data_path : pathlib.Path
            The path of the BIDS directory.
        """
        # Create the data path based on the available entities:
        # root, subject, session, and datatype
        data_path = "" if self.root is None else self.root
        if self.subject is not None:
            data_path = op.join(data_path, f"sub-{self.subject}")
        if self.session is not None:
            data_path = op.join(data_path, f"ses-{self.session}")
        # datatype will allow 'meg', 'eeg', 'ieeg', 'anat'
        if self.datatype is not None:
            data_path = op.join(data_path, self.datatype)
        return Path(data_path)

    @property
    def subject(self) -> Optional[str]:
        """The subject ID."""
        return self._subject

    @subject.setter
    def subject(self, value):
        self.update(subject=value)

    @property
    def session(self) -> Optional[str]:
        """The acquisition session."""
        return self._session

    @session.setter
    def session(self, value):
        self.update(session=value)

    @property
    def task(self) -> Optional[str]:
        """The experimental task."""
        return self._task

    @task.setter
    def task(self, value):
        self.update(task=value)

    @property
    def run(self) -> Optional[str]:
        """The run number."""
        return self._run

    @run.setter
    def run(self, value):
        self.update(run=value)

    @property
    def acquisition(self) -> Optional[str]:
        """The acquisition parameters."""
        return self._acquisition

    @acquisition.setter
    def acquisition(self, value):
        self.update(acquisition=value)

    @property
    def processing(self) -> Optional[str]:
        """The processing label."""
        return self._processing

    @processing.setter
    def processing(self, value):
        self.update(processing=value)

    @property
    def recording(self) -> Optional[str]:
        """The recording name."""
        return self._recording

    @recording.setter
    def recording(self, value):
        self.update(recording=value)

    @property
    def space(self) -> Optional[str]:
        """The coordinate space for an anatomical or sensor position file."""
        return self._space

    @space.setter
    def space(self, value):
        self.update(space=value)

    @property
    def description(self) -> Optional[str]:
        """The description entity."""
        return self._description

    @description.setter
    def description(self, value):
        self.update(description=value)

    @property
    def suffix(self) -> Optional[str]:
        """The filename suffix."""
        return self._suffix

    @suffix.setter
    def suffix(self, value):
        self.update(suffix=value)

    @property
    def root(self) -> Optional[Path]:
        """The root directory of the BIDS dataset."""
        return self._root

    @root.setter
    def root(self, value):
        self.update(root=value)

    @property
    def datatype(self) -> Optional[str]:
        """The BIDS data type, e.g. ``'anat'``, ``'meg'``, ``'eeg'``."""
        return self._datatype

    @datatype.setter
    def datatype(self, value):
        self.update(datatype=value)

    @property
    def split(self) -> Optional[str]:
        """The split of the continuous recording file for ``.fif`` data."""
        return self._split

    @split.setter
    def split(self, value):
        self.update(split=value)

    @property
    def extension(self) -> Optional[str]:
        """The extension of the filename, including a leading period."""
        return self._extension

    @extension.setter
    def extension(self, value):
        self.update(extension=value)

    def __str__(self):
        """Return the string representation of the path."""
        return str(self.fpath.as_posix())

    def __repr__(self):
        """Representation in the style of `pathlib.Path`."""
        root = self.root.as_posix() if self.root is not None else None

        return (
            f"{self.__class__.__name__}(\n"
            f"root: {root}\n"
            f"datatype: {self.datatype}\n"
            f"basename: {self.basename})"
        )

    def __fspath__(self):
        """Return the string representation for any fs functions."""
        return str(self.fpath)

    def __eq__(self, other):
        """Compare str representations."""
        return str(self) == str(other)

    def __ne__(self, other):
        """Compare str representations."""
        return str(self) != str(other)

    def copy(self):
        """Copy the instance.

        Returns
        -------
        bidspath : BIDSPath
            The copied bidspath.
        """
        return deepcopy(self)

    def mkdir(self, exist_ok=True):
        """Create the directory structure of the BIDS path.

        Parameters
        ----------
        exist_ok : bool
            If ``False``, raise an exception if the directory already exists.
            Otherwise, do nothing (default).

        Returns
        -------
        self : BIDSPath
            The BIDSPath object.
        """
        self.directory.mkdir(parents=True, exist_ok=exist_ok)
        return self

    @verbose
    def rm(self, *, safe_remove=True, verbose=None):
        """Safely delete a set of files from a BIDS dataset.

        Deleting a scan that conforms to the bids-validator will
        remove the respective row in ``*_scans.tsv``,  the
        corresponding sidecar files, and the data file itself.

        Deleting all files of a subject will update the
        ``*_participants.tsv`` file.


        Parameters
        ----------
        safe_remove : bool
            If ``False``, directly delete and update the files.
            Otherwise, displays the list of operations planned
            and asks for user confirmation before
            executing them (default).
        %(verbose)s

        Returns
        -------
        self : BIDSPath
            The BIDSPath object.

        Examples
        --------
        Remove one specific run:

        >>> bids_path = BIDSPath(subject='01', session='01', run="01",  # doctest: +SKIP
        ...                      root='/bids_dataset').rm()  # doctest: +SKIP
        Please, confirm you want to execute the following operations:
        Delete:
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_run-01_channels.tsv
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_run-01_events.json
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_run-01_events.tsv
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_run-01_meg.fif
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_run-01_meg.json
        Update:
        /bids_dataset/sub-01/ses-01/sub-01_ses-01_scans.tsv
        I confirm [y/N]>? y

        Remove all the files of a specific subject:

        >>> bids_path = BIDSPath(subject='01', root='/bids_dataset',  # doctest: +SKIP
        ...                      check=False).rm()  # doctest: +SKIP
        Please, confirm you want to execute the following operations:
        Delete:
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_acq-calibration_meg.dat
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_acq-crosstalk_meg.fif
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_coordsystem.json
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_run-02_channels.tsv
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_run-02_events.json
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_run-02_events.tsv
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_run-02_meg.fif
        /bids_dataset/sub-01/ses-01/meg/sub-01_ses-01_run-02_meg.json
        /bids_dataset/sub-01/ses-01/sub-01_ses-01_scans.tsv
        /bids_dataset/sub-01
        Update:
        /bids_dataset/participants.tsv
        I confirm [y/N]>? y
        """
        # only proceed if root is defined
        if self.root is None:
            raise RuntimeError("The root must not be None to remove files.")

        # Planning:
        paths_matched = self.match(ignore_json=False, check=self.check)
        subjects = set()
        paths_to_delete = list()
        paths_to_update = {}
        subjects_paths_to_delete = []
        participants_tsv_fpath = None
        for bids_path in paths_matched:
            paths_to_delete.append(bids_path)
            # if a datatype is present, then check
            # if a scan is deleted or not
            if bids_path.datatype is not None:
                # read in the corresponding scans file
                scans_fpath = (
                    bids_path.copy()
                    .update(datatype=None)
                    .find_matching_sidecar(
                        suffix="scans",
                        extension=".tsv",
                        on_error="raise",
                    )
                )
                paths_to_update.setdefault(scans_fpath, []).append(bids_path)
            subjects.add(bids_path.subject)

        files_to_delete = set(p.fpath for p in paths_to_delete)
        for subject in subjects:
            # check existence of files in the subject dir
            subj_path = BIDSPath(root=self.root, subject=subject)
            subj_files = [
                fpath for fpath in subj_path.directory.rglob("*") if fpath.is_file()
            ]
            if set(subj_files) <= files_to_delete:
                subjects_paths_to_delete.append(subj_path)
                participants_tsv_fpath = self.root / "participants.tsv"

        # Informing:
        pretty_delete_paths = "\n".join(
            [
                str(p)
                for p in paths_to_delete
                + [p.directory for p in subjects_paths_to_delete]
            ]
        )
        pretty_update_paths = "\n".join(
            [
                str(p)
                for p in list(paths_to_update.keys())
                + (
                    [participants_tsv_fpath]
                    if participants_tsv_fpath is not None
                    else []
                )
            ]
        )
        summary = ""
        if pretty_delete_paths:
            summary += f"Delete:\n{pretty_delete_paths}\n"
        if pretty_update_paths:
            summary += f"Update:\n{pretty_update_paths}\n"

        if safe_remove:
            choice = input(
                "Please, confirm you want to execute the following operations:\n"
                f"{summary}\nI confirm [y/N]"
            )
            if choice.lower() != "y":
                return
        else:
            logger.info(f"Executing the following operations:\n{summary}")

        # Execution:
        for bids_path in paths_to_delete:
            bids_path.fpath.unlink()

        for scans_fpath, bids_paths in paths_to_update.items():
            if not scans_fpath.exists():
                continue
            # get the relative datatype of these bids files
            bids_fnames = [op.join(p.datatype, p.fpath.name) for p in bids_paths]

            scans_tsv = _from_tsv(scans_fpath)
            scans_tsv = _drop(scans_tsv, bids_fnames, "filename")
            _to_tsv(scans_tsv, scans_fpath)

        subjects_to_delete = []
        for subj_path in subjects_paths_to_delete:
            if subj_path.directory.exists():
                sh.rmtree(subj_path.directory)
            subjects_to_delete.append(subj_path.subject)
        if subjects_to_delete and participants_tsv_fpath.exists():
            participants_tsv = _from_tsv(participants_tsv_fpath)
            participants_tsv = _drop(
                participants_tsv, subjects_to_delete, "participant_id"
            )
            _to_tsv(participants_tsv, participants_tsv_fpath)

        return self

    @property
    def fpath(self):
        """Full filepath for this BIDS file.

        Getting the file path consists of the entities passed in
        and will get the relative (or full if ``root`` is passed)
        path.

        Returns
        -------
        bids_fpath : pathlib.Path
            Either the relative, or full path to the dataset.
        """
        # get the inner-most BIDS directory for this file path
        data_path = self.directory

        # account for MEG data that are directory-based
        # else, all other file paths attempt to match
        if self.suffix == "meg" and self.extension == ".ds":
            bids_fpath = op.join(data_path, self.basename)
        elif self.suffix == "meg" and self.extension == ".pdf":
            bids_fpath = op.join(data_path, op.splitext(self.basename)[0])
        else:
            # if suffix and/or extension is missing, and root is
            # not None, then BIDSPath will infer the dataset
            # else, return the relative path with the basename
            if (
                self.suffix is None or self.extension is None
            ) and self.root is not None:
                # get matching BIDSPaths inside the bids root
                matching_paths = _get_matching_bidspaths_from_filesystem(self)

                # FIXME This will break
                # FIXME e.g. with FIFF data split across multiple files.
                # if extension is not specified and no unique file path
                # return filepath of the actual dataset for MEG/EEG/iEEG data
                if self.suffix is None or self.suffix in ALLOWED_DATATYPES:
                    # now only use valid datatype extension
                    if self.extension is None:
                        valid_exts = sum(ALLOWED_DATATYPE_EXTENSIONS.values(), [])
                    else:
                        valid_exts = [self.extension]
                    matching_paths = [
                        p for p in matching_paths if _parse_ext(p)[1] in valid_exts
                    ]

                if self.split is None and (
                    not matching_paths or "_split-" in matching_paths[0]
                ):
                    # try finding FIF split files (only first one)
                    this_self = self.copy().update(split="01")
                    matching_paths = _get_matching_bidspaths_from_filesystem(this_self)

                # found no matching paths
                if not matching_paths:
                    bids_fpath = op.join(data_path, self.basename)
                # if paths still cannot be resolved, then there is an error
                elif len(matching_paths) > 1:
                    matching_paths_str = "\n".join(sorted(matching_paths))
                    msg = (
                        "Found more than one matching data file for the "
                        "requested recording. While searching:\n"
                        f'{indent(repr(self), "    ")}\n'
                        f"Found {len(matching_paths)} paths:\n"
                        f'{indent(matching_paths_str, "    ")}\n'
                        "Cannot proceed due to the "
                        "ambiguity. This is likely a problem with your "
                        "BIDS dataset. Please run the BIDS validator on "
                        "your data."
                    )
                    raise RuntimeError(msg)
                else:
                    bids_fpath = matching_paths[0]

            else:
                bids_fpath = op.join(data_path, self.basename)

        bids_fpath = Path(bids_fpath)
        return bids_fpath

    def update(self, *, check=None, **kwargs):
        """Update inplace BIDS entity key/value pairs in object.

        ``run`` and ``split`` are auto-converted to have two
        digits. For example, if ``run=1``, then it will nbecome ``run='01'``.

        Also performs error checks on various entities to
        adhere to the BIDS specification. Specifically:
        - ``datatype`` should be one of: ``anat``, ``eeg``, ``ieeg``, ``meg``
        - ``extension`` should be one of the accepted file
        extensions in the file path: ``.con``, ``.sqd``, ``.fif``,
        ``.pdf``, ``.ds``, ``.vhdr``, ``.edf``, ``.bdf``, ``.set``,
        ``.edf``, ``.set``, ``.mef``, ``.nwb``
        - ``suffix`` should be one of the acceptable file suffixes in: ``meg``,
        ``markers``, ``eeg``, ``ieeg``, ``T1w``,
        ``participants``, ``scans``, ``electrodes``, ``channels``,
        ``coordsystem``, ``events``, ``headshape``, ``digitizer``,
        ``beh``, ``physio``, ``stim``
        - Depending on the modality of the data (EEG, MEG, iEEG),
        ``space`` should be a valid string according to Appendix VIII
        in the BIDS specification.

        Parameters
        ----------
        check : None | bool
            If a boolean, controls whether to enforce BIDS conformity. This
            will set the ``.check`` attribute accordingly. If ``None``, rely on
            the existing ``.check`` attribute instead, which is set upon
            `mne_bids.BIDSPath` instantiation. Defaults to ``None``.
        **kwargs : dict
            It can contain updates for valid BIDSPath entities:
            'subject', 'session', 'task', 'acquisition', 'processing', 'run',
            'recording', 'space', 'suffix', 'split', 'extension',
            or updates for 'root' or 'datatype'.

        Returns
        -------
        bidspath : BIDSPath
            The updated instance of BIDSPath.

        Examples
        --------
        If one creates a bids basename using
        :func:`mne_bids.BIDSPath`:

        >>> bids_path = BIDSPath(subject='test', session='two',
        ...                      task='mytask', suffix='channels',
        ...                      extension='.tsv')
        >>> print(bids_path.basename)
        sub-test_ses-two_task-mytask_channels.tsv
        >>> # Then, one can update this `BIDSPath` object in place
        >>> bids_path.update(acquisition='test', suffix='ieeg',
        ...                  datatype='ieeg',
        ...                  extension='.vhdr', task=None)
        BIDSPath(
        root: None
        datatype: ieeg
        basename: sub-test_ses-two_acq-test_ieeg.vhdr)
        >>> print(bids_path.basename)
        sub-test_ses-two_acq-test_ieeg.vhdr
        """
        # Update .check attribute
        if check is not None:
            self.check = check

        for key, val in kwargs.items():
            if key == "root":
                _validate_type(val, types=("path-like", None), item_name=key)
                continue

            if key == "datatype":
                if val is not None and val not in ALLOWED_DATATYPES and self.check:
                    raise ValueError(
                        f"datatype ({val}) is not valid. "
                        f"Should be one of "
                        f"{ALLOWED_DATATYPES}"
                    )
                else:
                    continue

            if key not in ENTITY_VALUE_TYPE:
                raise ValueError(
                    f"Key must be one of " f"{ALLOWED_PATH_ENTITIES}, got {key}"
                )

            if ENTITY_VALUE_TYPE[key] == "label":
                _validate_type(val, types=(None, str), item_name=key)
            else:
                assert ENTITY_VALUE_TYPE[key] == "index"
                _validate_type(val, types=(int, str, None), item_name=key)
                if isinstance(val, str) and not val.isdigit():
                    raise ValueError(f"{key} is not an index (Got {val})")
                elif isinstance(val, int):
                    kwargs[key] = f"{val}"

        # ensure extension starts with a '.'
        extension = kwargs.get("extension")
        if extension is not None and not extension.startswith("."):
            kwargs["extension"] = f".{extension}"
        del extension

        # error check entities
        old_kwargs = dict()
        for key, val in kwargs.items():
            # check if there are any characters not allowed
            if val is not None and key != "root":
                if key == "suffix" and not self.check:
                    # suffix may skip a check if check=False to allow
                    # things like "dataset_description.json"
                    pass
                else:
                    _check_key_val(key, val)

            # set entity value, ensuring `root` is a Path
            if val is not None and key == "root":
                val = Path(val).expanduser()
            old_kwargs[key] = (
                getattr(self, f"{key}") if hasattr(self, f"_{key}") else None
            )
            setattr(self, f"_{key}", val)

        # Perform a check of the entities and revert changes if check fails
        try:
            self._check()
        except Exception as e:
            old_check = self.check
            self.check = False
            self.update(**old_kwargs)
            self.check = old_check
            raise e
        return self

    def match(self, ignore_json=True, check=False):
        """Get a list of all matching paths in the root directory.

        Performs a recursive search, starting in ``.root`` (if set), based on
        `BIDSPath.entities` object. Ignores ``.json`` files.

        Parameters
        ----------
        ignore_json : bool
            If ``True``, ignores json files. Defaults to ``True``.
        check : bool
            If ``True``, only returns paths that conform to BIDS. If ``False``
            (default), the ``.check`` attribute of the returned
            `mne_bids.BIDSPath` object will be set to ``True`` for paths that
            do conform to BIDS, and to ``False`` for those that don't.

        Returns
        -------
        bids_paths : list of mne_bids.BIDSPath
            The matching paths.
        """
        if self.root is None:
            raise RuntimeError(
                "Cannot match basenames if `root` "
                "attribute is not set. Please set the"
                "BIDS root directory path to `root` via "
                "BIDSPath.update()."
            )

        paths = _return_root_paths(
            self.root, datatype=self.datatype, ignore_json=ignore_json
        )

        fnames = _filter_fnames(
            paths, suffix=self.suffix, extension=self.extension, **self.entities
        )

        bids_paths = _fnames_to_bidspaths(fnames, self.root, check=check)
        return bids_paths

    def _check(self):
        """Deep check or not of the instance."""
        self.basename  # run basename to check validity of arguments

        # perform error check on scans
        if (
            self.suffix == "scans" and self.extension == ".tsv"
        ) and _check_non_sub_ses_entity(self):
            raise ValueError(
                "scans.tsv file name can only contain "
                "subject and session entities. BIDSPath "
                f"currently contains {self.entities}."
            )

        # perform deeper check if user has it turned on
        if self.check:
            _check_empty_room_basename(self)

            if (
                self.acquisition in ("calibration", "crosstalk")
                and self.task is not None
            ):
                raise ValueError(
                    f'task must be None if the acquisition is "calibration" or '
                    f'"crosstalk", but received: {self.task}'
                )

            # ensure extension starts with a '.'
            extension = self.extension
            if extension is not None:
                # check validity of the extension
                if extension not in ALLOWED_FILENAME_EXTENSIONS:
                    raise ValueError(
                        f"Extension {extension} is not "
                        f"allowed. Use one of these extensions "
                        f"{ALLOWED_FILENAME_EXTENSIONS}."
                    )

            # labels from space entity must come from list (appendix VIII)
            space = self.space
            if space is not None:
                datatype = getattr(self, "datatype", None)
                if datatype is None:
                    raise ValueError(
                        "You must define datatype if you want to "
                        "use space in your BIDSPath."
                    )

                allowed_spaces_for_dtype = ALLOWED_SPACES.get(datatype, None)
                if allowed_spaces_for_dtype is None:
                    raise ValueError(
                        f"space entity is not valid for datatype " f"{self.datatype}"
                    )
                elif space not in allowed_spaces_for_dtype:
                    raise ValueError(
                        f"space ({space}) is not valid for "
                        f"datatype ({self.datatype}).\n"
                        f"Should be one of "
                        f"{allowed_spaces_for_dtype}"
                    )
                else:
                    pass

            # error check suffix
            suffix = self.suffix
            if suffix is not None and suffix not in ALLOWED_FILENAME_SUFFIX:
                raise ValueError(
                    f"Suffix {suffix} is not allowed. "
                    f"Use one of these suffixes "
                    f"{ALLOWED_FILENAME_SUFFIX}."
                )

    @verbose
    def find_empty_room(self, use_sidecar_only=False, *, verbose=None):
        """Find the corresponding empty-room file of an MEG recording.

        This will only work if the ``.root`` attribute of the
        :class:`mne_bids.BIDSPath` instance has been set.

        Parameters
        ----------
        use_sidecar_only : bool
            Whether to only check the ``AssociatedEmptyRoom`` entry in the
            sidecar JSON file or not. If ``False``, first look for the entry,
            and if unsuccessful, try to find the best-matching empty-room
            recording in the dataset based on the measurement date.
        %(verbose)s

        Returns
        -------
        BIDSPath | None
            The path corresponding to the best-matching empty-room measurement.
            Returns ``None`` if none was found.
        """
        if self.datatype not in ("meg", None):
            raise ValueError("Empty-room data is only supported for MEG datasets")

        if self.root is None:
            raise ValueError(
                'The root of the "bids_path" must be set. '
                'Please use `bids_path.update(root="<root>")` '
                "to set the root of the BIDS folder to read."
            )

        # needed to deal with inheritance principle
        sidecar_fname = (
            self.copy()
            .update(datatype=None, suffix="meg")
            .find_matching_sidecar(extension=".json")
        )
        with open(sidecar_fname, encoding="utf-8") as f:
            sidecar_json = json.load(f)

        if "AssociatedEmptyRoom" in sidecar_json:
            logger.info(
                'Using "AssociatedEmptyRoom" entry from MEG sidecar '
                "file to retrieve empty-room path."
            )
            emptytoom_path = sidecar_json["AssociatedEmptyRoom"]
            er_bids_path = get_bids_path_from_fname(emptytoom_path)
            er_bids_path.root = self.root
            er_bids_path.datatype = "meg"
        elif use_sidecar_only:
            logger.info(
                "The MEG sidecar file does not contain an "
                '"AssociatedEmptyRoom" entry. Aborting search for an '
                "empty-room recording, as you passed use_sidecar_only=True"
            )
            return None
        else:
            logger.info(
                "The MEG sidecar file does not contain an "
                '"AssociatedEmptyRoom" entry. Will try to find a matching '
                "empty-room recording based on the measurement date …"
            )
            er_bids_path = _find_matched_empty_room(self)

        if er_bids_path is not None and not er_bids_path.fpath.exists():
            raise FileNotFoundError(
                f"Empty-room BIDS path resolved but not found:\n"
                f"{er_bids_path}\n"
                "Check your BIDS dataset for completeness."
            )

        return er_bids_path

    def get_empty_room_candidates(self):
        """Get the list of empty-room candidates for the given file.

        Returns
        -------
        candidates : list of BIDSPath
            The candidate files that will be checked if the sidecar does not
            contain an "AssociatedEmptyRoom" entry.

        Notes
        -----
        .. versionadded:: 0.12.0
        """
        return _find_empty_room_candidates(self)

    def find_matching_sidecar(self, suffix=None, extension=None, *, on_error="raise"):
        """Get the matching sidecar JSON path.

        Parameters
        ----------
        suffix : str | None
            The filename suffix. This is the entity after the last ``_``
            before the extension. E.g., ``'ieeg'``.
        extension : str | None
            The extension of the filename. E.g., ``'.json'``.
        on_error : 'raise' | 'warn' | 'ignore'
            If no matching sidecar file was found and this is set to
            ``'raise'``, raise a ``RuntimeError``. If ``'warn'``, emit a
            warning, and if ``'ignore'``, neither raise an exception nor a
            warning, and return ``None`` in both cases.

        Returns
        -------
        sidecar_path : pathlib.Path | None
            The path to the sidecar JSON file.
        """
        return _find_matching_sidecar(
            self,
            suffix=suffix,
            extension=extension,
            on_error=on_error,
        )

    @property
    def meg_calibration_fpath(self):
        """Find the matching Elekta/Neuromag/MEGIN fine-calibration file.

        This requires that at least ``root`` and ``subject`` are set, and that
        ``datatype`` is either ``'meg'`` or ``None``.

        Returns
        -------
        path : pathlib.Path | None
            The path of the fine-calibration file, or ``None`` if it couldn't
            be found.
        """
        if self.root is None or self.subject is None:
            raise ValueError("root and subject must be set.")
        if self.datatype not in (None, "meg"):
            raise ValueError("Can only find fine-calibration file for MEG datasets.")

        path = BIDSPath(
            subject=self.subject,
            session=self.session,
            acquisition="calibration",
            suffix="meg",
            extension=".dat",
            datatype="meg",
            root=self.root,
        ).fpath
        if not path.exists():
            path = None

        return path

    @property
    def meg_crosstalk_fpath(self):
        """Find the matching Elekta/Neuromag/MEGIN crosstalk file.

        This requires that at least ``root`` and ``subject`` are set, and that
        ``datatype`` is either ``'meg'`` or ``None``.

        Returns
        -------
        path : pathlib.Path | None
            The path of the crosstalk file, or ``None`` if it couldn't be
            found.
        """
        if self.root is None or self.subject is None:
            raise ValueError("root and subject must be set.")
        if self.datatype not in (None, "meg"):
            raise ValueError("Can only find crosstalk file for MEG datasets.")

        path = BIDSPath(
            subject=self.subject,
            session=self.session,
            acquisition="crosstalk",
            suffix="meg",
            extension=".fif",
            datatype="meg",
            root=self.root,
        ).fpath
        if not path.exists():
            path = None

        return path


def _get_matching_bidspaths_from_filesystem(bids_path):
    """Get matching file paths for a BIDSPath.

    Assumes suffix and/or extension is not provided.
    """
    # extract relevant entities to find filepath
    sub, ses = bids_path.subject, bids_path.session
    datatype = bids_path.datatype
    basename, bids_root = bids_path.basename, bids_path.root

    if datatype is None:
        datatype = _infer_datatype(root=bids_root, sub=sub, ses=ses)

    data_dir = BIDSPath(
        subject=sub, session=ses, datatype=datatype, root=bids_root
    ).directory

    # For BTi data, just return the directory with a '.pdf' extension
    # to facilitate reading in mne-bids
    bti_dir = op.join(data_dir, f"{basename}")
    if op.isdir(bti_dir):
        logger.info(f"Assuming BTi data in {bti_dir}")
        matching_paths = [f"{bti_dir}.pdf"]
    # otherwise, search for valid file paths
    else:
        search_str = bids_root
        # parse down the BIDS directory structure
        if sub is not None:
            search_str = op.join(search_str, f"sub-{sub}")
        if ses is not None:
            search_str = op.join(search_str, f"ses-{ses}")
        if datatype is not None:
            search_str = op.join(search_str, datatype)
        else:
            search_str = op.join(search_str, "**")
        search_str = op.join(search_str, f"{basename}*")

        # Find all matching files in all supported formats.
        valid_exts = ALLOWED_FILENAME_EXTENSIONS
        matching_paths = glob.glob(search_str)
        matching_paths = [p for p in matching_paths if _parse_ext(p)[1] in valid_exts]
    return matching_paths


def _check_non_sub_ses_entity(bids_path):
    """Check existence of non subject/session entities in BIDSPath."""
    if (
        bids_path.task
        or bids_path.acquisition
        or bids_path.run
        or bids_path.space
        or bids_path.recording
        or bids_path.split
        or bids_path.processing
    ):
        return True
    return False


def _print_lines_with_entry(file, entry, folder, is_tsv, line_numbers, outfile):
    """Print the lines that contain the entry.

    Parameters
    ----------
    file : str
        The text file to look though.
    entry : str
        The string to look in the text file for.
    folder : str
        The base folder for relative file path printing.
    is_tsv : bool
        If ``True``, things that format a tsv nice will be used.
    line_numbers : bool
        Whether to include line numbers in the printout.
    outfile : io.StringIO | None
        The argument to pass to `print` for `file`. If ``None``,
        prints to the console, else a string is printed to.
    """
    entry_lines = list()
    with open(file, encoding="utf-8-sig") as fid:
        if is_tsv:  # format tsv files nicely
            header = _truncate_tsv_line(fid.readline())
            if line_numbers:
                header = f"1    {header}"
            header = header.rstrip()
        for i, line in enumerate(fid):
            if entry in line:
                if is_tsv:
                    line = _truncate_tsv_line(line)
                if line_numbers:
                    line = str(i + 2) + (5 - len(str(i + 2))) * " " + line
                entry_lines.append(line.rstrip())
    if entry_lines:
        print(op.relpath(file, folder), file=outfile)
        if is_tsv:
            print(f"    {header}", file=outfile)
        if len(entry_lines) > 10:
            entry_lines = entry_lines[:10]
            entry_lines.append("...")
        for line in entry_lines:
            print(f"    {line}", file=outfile)


def _truncate_tsv_line(line, lim=10):
    """Truncate a line to the specified number of characters."""
    return "".join(
        [
            str(val) + (lim - len(val)) * " " if len(val) < lim else f"{val[:lim - 1]} "
            for val in line.split("\t")
        ]
    )


def search_folder_for_text(
    entry, folder, extensions=(".json", ".tsv"), line_numbers=True, return_str=False
):
    """Find any particular string entry in the text files of a folder.

    .. note:: This is a search function like `grep
              <https://man7.org/linux/man-pages/man1/fgrep.1.html>`_
              that is formatted nicely for BIDS datasets.

    Parameters
    ----------
    entry : str
        The string to search for
    folder : path-like
        The folder in which to search.
    extensions : list | tuple | str
        The extensions to search through. Default is ``json`` and
        ``tsv`` which are the BIDS sidecar file types.
    line_numbers : bool
        Whether to include line numbers.
    return_str : bool
        If ``True``, return the fields with "n/a" as a str instead of
        printing them.

    Returns
    -------
    str | None
        If `return_str` is ``True``, the fields are returned as a
        string. Else, ``None`` is returned and the fields are printed.
    """
    _validate_type(entry, str, "entry")
    if not op.isdir(folder):
        raise ValueError("{folder} is not a directory")
    folder = Path(folder)  # ensure pathlib.Path

    extensions = (extensions,) if isinstance(extensions, str) else extensions
    _validate_type(extensions, (tuple, list))
    _validate_type(line_numbers, bool, "line_numbers")
    _validate_type(return_str, bool, "return_str")
    outfile = StringIO() if return_str else None

    for extension in extensions:
        for file in folder.rglob("*" + extension):
            _print_lines_with_entry(
                file, entry, folder, extension == ".tsv", line_numbers, outfile
            )

    if outfile is not None:
        return outfile.getvalue()


def _check_max_depth(max_depth):
    """Check that max depth is a proper input."""
    msg = "`max_depth` must be a positive integer or None"
    if not isinstance(max_depth, (int, type(None))):
        raise ValueError(msg)
    if max_depth is None:
        max_depth = float("inf")
    if max_depth < 0:
        raise ValueError(msg)
    # Use max_depth same as the -L param in the unix `tree` command
    max_depth += 1
    return max_depth


def print_dir_tree(folder, max_depth=None, return_str=False):
    """Recursively print a directory tree.

    Parameters
    ----------
    folder : path-like
        The folder for which to print the directory tree.
    max_depth : int
        The maximum depth into which to descend recursively for printing
        the directory tree.
    return_str : bool
        If ``True``, return the directory tree as a str instead of
        printing it.

    Returns
    -------
    str | None
        If `return_str` is ``True``, the directory tree is returned as a
        string. Else, ``None`` is returned and the directory tree is printed.
    """
    folder = _check_fname(
        fname=folder, overwrite="read", must_exist=True, name="Folder", need_dir=True
    )
    max_depth = _check_max_depth(max_depth)

    _validate_type(return_str, bool, "return_str")
    outfile = StringIO() if return_str else None

    # Base length of a tree branch, to normalize each tree's start to 0
    baselen = len(str(folder).split(os.sep)) - 1

    # Recursively walk through all directories
    for root, dirs, files in os.walk(folder, topdown=True):
        # Since we're using `topdown=True`, sorting `dirs` ensures that
        # `os.walk` will continue walking through directories in alphabetical
        # order. So although we're not actually using `dirs` anywhere below,
        # sorting it here is imperative to ensure the correct (alphabetical)
        # directory sort order in the output.
        dirs.sort()
        files.sort()

        # Check how far we have walked
        branchlen = len(root.split(os.sep)) - baselen

        # Only print if this is up to the depth we asked
        if branchlen <= max_depth:
            if branchlen <= 1:
                print(f"|{op.basename(root) + os.sep}", file=outfile)
            else:
                print(
                    "|{} {}".format(
                        (branchlen - 1) * "---", op.basename(root) + os.sep
                    ),
                    file=outfile,
                )

            # Only print files if we are NOT yet up to max_depth or beyond
            if branchlen < max_depth:
                for file in files:
                    print("|{} {}".format(branchlen * "---", file), file=outfile)

    if outfile is not None:
        return outfile.getvalue()


def _parse_ext(raw_fname):
    """Split a filename into its name and extension."""
    raw_fname = str(raw_fname)
    fname, ext = os.path.splitext(raw_fname)
    # BTi data is the only file format that does not have a file extension
    if ext == "" or "c,rf" in fname:
        logger.info(
            'Found no extension for raw file, assuming "BTi" format '
            "and appending extension .pdf"
        )
        ext = ".pdf"
    # If ending on .gz, check whether it is an .nii.gz file
    elif ext == ".gz" and raw_fname.endswith(".nii.gz"):
        ext = ".nii.gz"
        fname = fname[:-4]  # cut off the .nii
    return fname, ext


def _infer_datatype_from_path(fname: Path):
    # get the parent
    if fname.exists():
        datatype = fname.parent.name
        if any([datatype.startswith(entity) for entity in ["sub", "ses"]]):
            datatype = None
    elif fname.stem.split("_")[-1] in ("meg", "eeg", "ieeg"):
        datatype = fname.stem.split("_")[-1]
    else:
        datatype = None

    return datatype


@verbose
def get_bids_path_from_fname(fname, check=True, verbose=None):
    """Retrieve a BIDSPath object from a filename.

    Parameters
    ----------
    fname : path-like
        The path to parse a `BIDSPath` from.
    check : bool
        Whether to check if the generated `BIDSPath` complies with the BIDS
        specification, i.e., whether all included entities and the suffix are
        valid.
    %(verbose)s

    Returns
    -------
    bids_path : BIDSPath
        The BIDSPath object.
    """
    fpath = Path(fname)
    fname = fpath.name

    entities = get_entities_from_fname(fname)

    # parse suffix and extension
    last_entity = fname.split("-")[-1]
    if "_" in last_entity:
        suffix = last_entity.split("_")[-1]
        suffix, extension = _get_bids_suffix_and_ext(suffix)
    else:
        suffix = None
        extension = Path(fname).suffix  # already starts with a period
        if extension == "":
            extension = None

    if extension is not None:
        assert extension.startswith(".")  # better safe than sorry

    datatype = _infer_datatype_from_path(fpath)

    # find root and datatype if it exists
    if fpath.parent == "":
        root = None
    else:
        root_level = 0
        # determine root if it's there
        if entities["subject"] is not None:
            root_level += 1
        if entities["session"] is not None:
            root_level += 1
        if suffix != "scans":
            root_level += 1

        if root_level:
            root = fpath.parent
            for _ in range(root_level):
                root = root.parent

    bids_path = BIDSPath(
        root=root,
        datatype=datatype,
        suffix=suffix,
        extension=extension,
        **entities,
        check=check,
    )
    if verbose:
        logger.info(f"From {fpath}, formed a BIDSPath: {bids_path}.")
    return bids_path


@verbose
def get_entities_from_fname(fname, on_error="raise", verbose=None):
    """Retrieve a dictionary of BIDS entities from a filename.

    Entities not present in ``fname`` will be assigned the value of ``None``.

    Parameters
    ----------
    fname : BIDSPath | path-like
        The path to parse.
    on_error : 'raise' | 'warn' | 'ignore'
        If any unsupported labels in the filename are found and this is set
        to ``'raise'``, raise a ``RuntimeError``. If ``'warn'``,
        emit a warning and continue, and if ``'ignore'``,
        neither raise an exception nor a warning, and
        return all entities found. For example, currently MNE-BIDS does not
        support derivatives yet, but the ``desc`` entity label is used to
        differentiate different derivatives and will work with this function
        if ``on_error='ignore'``.
    %(verbose)s

    Returns
    -------
    params : dict
        A dictionary with the keys corresponding to the BIDS entity names, and
        the values to the entity values encoded in the filename.

    Examples
    --------
    >>> fname = 'sub-01_ses-exp_run-02_meg.fif'
    >>> get_entities_from_fname(fname)
    {'subject': '01', \
'session': 'exp', \
'task': None, \
'acquisition': None, \
'run': '02', \
'processing': None, \
'space': None, \
'recording': None, \
'split': None, \
'description': None}
    """
    if on_error not in ("warn", "raise", "ignore"):
        raise ValueError(
            f"Acceptable values for on_error are: warn, raise, "
            f"ignore, but got: {on_error}"
        )

    fname = str(fname)  # to accept also BIDSPath or Path instances

    # filename keywords to the BIDS entity mapping
    entity_vals = list(ALLOWED_PATH_ENTITIES_SHORT.values())
    fname_vals = list(ALLOWED_PATH_ENTITIES_SHORT.keys())

    params = {key: None for key in entity_vals}
    idx_key = 0
    for match in re.finditer(param_regex, op.basename(fname)):
        key, value = match.groups()

        if on_error in ("raise", "warn"):
            if key not in fname_vals:
                msg = f'Unexpected entity "{key}" found in ' f'filename "{fname}"'
                if on_error == "raise":
                    raise KeyError(msg)
                elif on_error == "warn":
                    warn(msg)
                    continue
            if fname_vals.index(key) < idx_key:
                msg = (
                    f"Entities in filename not ordered correctly."
                    f' "{key}" should have occurred earlier in the '
                    f'filename "{fname}"'
                )
                raise ValueError(msg)
            idx_key = fname_vals.index(key)

        key_short_hand = ALLOWED_PATH_ENTITIES_SHORT.get(key, key)
        params[key_short_hand] = value
    return params


def _find_matching_sidecar(bids_path, suffix=None, extension=None, on_error="raise"):
    """Try to find a sidecar file with a given suffix for a data file.

    Parameters
    ----------
    bids_path : BIDSPath
        Full name of the data file.
    suffix : str | None
        The filename suffix. This is the entity after the last ``_``
        before the extension. E.g., ``'ieeg'``.
    extension : str | None
        The extension of the filename. E.g., ``'.json'``.
    on_error : 'raise' | 'warn' | 'ignore'
        If no matching sidecar file was found and this is set to ``'raise'``,
        raise a ``RuntimeError``. If ``'warn'``, emit a warning, and if
        ``'ignore'``, neither raise an exception nor a warning, and return
        ``None`` in both cases.

    Returns
    -------
    sidecar_fname : str | None
        Path to the identified sidecar file, or ``None`` if none could be found
        and ``on_error`` was set to ``'warn'`` or ``'ignore'``.

    """
    if on_error not in ("warn", "raise", "ignore"):
        raise ValueError(
            f"Acceptable values for on_error are: warn, raise, "
            f"ignore, but got: {on_error}"
        )

    bids_root = bids_path.root

    # search suffix is BIDS-suffix and extension
    search_suffix = ""
    if suffix is None and bids_path.suffix is not None:
        search_suffix = bids_path.suffix
    elif suffix is not None:
        search_suffix = suffix

        # do not search for suffix if suffix is explicitly passed
        bids_path = bids_path.copy()
        bids_path.check = False
        bids_path.update(suffix=None)

    if extension is None and bids_path.extension is not None:
        search_suffix = search_suffix + bids_path.extension
    elif extension is not None:
        search_suffix = search_suffix + extension

        # do not search for extension if extension is explicitly passed
        bids_path = bids_path.copy()
        bids_path.check = False
        bids_path = bids_path.update(extension=None)

    # We only use subject and session as identifier, because all other
    # parameters are potentially not binding for metadata sidecar files
    search_str_filename = f"sub-{bids_path.subject}"
    if bids_path.session is not None:
        search_str_filename += f"_ses-{bids_path.session}"

    # Find all potential sidecar files, doing a recursive glob
    # from bids_root/sub-*, potentially taking into account the data type
    search_dir = Path(bids_root) / f"sub-{bids_path.subject}"
    # ** -> don't forget about potentially present session directories
    if bids_path.datatype is None:
        search_dir = search_dir / "**"
    else:
        search_dir = search_dir / "**" / bids_path.datatype

    search_str_complete = str(search_dir / f"{search_str_filename}*{search_suffix}")

    candidate_list = glob.glob(search_str_complete, recursive=True)
    best_candidates = _find_best_candidates(bids_path.entities, candidate_list)
    if len(best_candidates) == 1:
        # Success
        return Path(best_candidates[0])

    # We failed. Construct a helpful error message.
    # If this was expected, simply return None, otherwise, raise an exception.
    msg = None
    if len(best_candidates) == 0:
        msg = (
            f"Did not find any {search_suffix} "
            f"associated with {bids_path.basename}."
        )
    elif len(best_candidates) > 1:
        # More than one candidates were tied for best match
        msg = (
            f"Expected to find a single {search_suffix} file "
            f"associated with {bids_path.basename}, "
            f"but found {len(candidate_list)}:\n\n" + "\n".join(candidate_list)
        )
    msg += f'\n\nThe search_str was "{search_str_complete}"'
    if on_error == "raise":
        raise RuntimeError(msg)
    elif on_error == "warn":
        warn(msg)

    return None


def _get_bids_suffix_and_ext(str_suffix):
    """Parse suffix for valid suffix and ext."""
    # no matter what the suffix is, suffix and extension are last
    suffix = str_suffix
    ext = None
    if "." in str_suffix:
        # handle case of multiple '.' in extension
        split_str = str_suffix.split(".")
        suffix = split_str[0]
        ext = ".".join(split_str[1:])
        ext = f".{ext}"  # prepend period
    return suffix, ext


@verbose
def get_datatypes(root, verbose=None):
    """Get list of data types ("modalities") present in a BIDS dataset.

    Parameters
    ----------
    root : path-like
        Path to the root of the BIDS directory.
    %(verbose)s

    Returns
    -------
    modalities : list of str
        List of the data types present in the BIDS dataset pointed to by
        `root`.

    """
    # Take all possible data types from "entity" table
    # (Appendix in BIDS spec)
    # https://bids-specification.readthedocs.io/en/latest/appendices/entity-table.html  # noqa
    datatype_list = ("anat", "func", "dwi", "fmap", "beh", "meg", "eeg", "ieeg", "nirs")
    datatypes = list()
    for root, dirs, files in os.walk(root):
        for dir in dirs:
            if dir in datatype_list and dir not in datatypes:
                datatypes.append(dir)

    return datatypes


@verbose
def get_entity_vals(
    root,
    entity_key,
    *,
    ignore_subjects="emptyroom",
    ignore_sessions=None,
    ignore_tasks=None,
    ignore_acquisitions=None,
    ignore_runs=None,
    ignore_processings=None,
    ignore_spaces=None,
    ignore_recordings=None,
    ignore_splits=None,
    ignore_descriptions=None,
    ignore_modalities=None,
    ignore_datatypes=None,
    ignore_dirs=("derivatives", "sourcedata"),
    with_key=False,
    verbose=None,
):
    """Get list of values associated with an `entity_key` in a BIDS dataset.

    BIDS file names are organized by key-value pairs called "entities" [1]_.
    With this function, you can get all values for an entity indexed by its
    key.

    Parameters
    ----------
    root : path-like
        Path to the "root" directory from which to start traversing to gather
        BIDS entities from file- and folder names. This will commonly be the
        BIDS root, but it may also be a subdirectory inside of a BIDS dataset,
        e.g., the ``sub-X`` directory of a hypothetical subject ``X``.

        .. note:: This function searches the names of all files and directories
                  nested within ``root``. Depending on the size of your
                  dataset and storage system, searching the entire BIDS dataset
                  may take a **considerable** amount of time (seconds up to
                  several minutes). If you find yourself running into such
                  performance issues, consider limiting the search to only a
                  subdirectory in the dataset, e.g., to a single subject or
                  session only.

    entity_key : str
        The name of the entity key to search for.
    ignore_subjects : str | array-like of str | None
        Subject(s) to ignore. By default, entities from the ``emptyroom``
        mock-subject are not returned. If ``None``, include all subjects.
    ignore_sessions : str | array-like of str | None
        Session(s) to ignore. If ``None``, include all sessions.
    ignore_tasks : str | array-like of str | None
        Task(s) to ignore. If ``None``, include all tasks.
    ignore_acquisitions : str | array-like of str | None
        Acquisition(s) to ignore. If ``None``, include all acquisitions.
    ignore_runs : str | array-like of str | None
        Run(s) to ignore. If ``None``, include all runs.
    ignore_processings : str | array-like of str | None
        Processing(s) to ignore. If ``None``, include all processings.
    ignore_spaces : str | array-like of str | None
        Space(s) to ignore. If ``None``, include all spaces.
    ignore_recordings : str | array-like of str | None
        Recording(s) to ignore. If ``None``, include all recordings.
    ignore_splits : str | array-like of str | None
        Split(s) to ignore. If ``None``, include all splits.
    ignore_descriptions : str | array-like of str | None
        Description(s) to ignore. If ``None``, include all descriptions.

        .. versionadded:: 0.11
    ignore_modalities : str | array-like of str | None
        Modalities(s) to ignore. If ``None``, include all modalities.
    ignore_datatypes : str | array-like of str | None
        Datatype(s) to ignore. If ``None``, include all datatypes (i.e.
        ``anat``, ``ieeg``, ``eeg``, ``meg``, ``func``, etc.)
    ignore_dirs : str | array-like of str | None
        Directories nested directly within ``root`` to ignore. If ``None``,
        include all directories in the search.

        .. versionadded:: 0.9
    with_key : bool
        If ``True``, returns the full entity with the key and the value. This
        will for example look like ``['sub-001', 'sub-002']``.
        If ``False`` (default), just returns the entity values. This
        will for example look like ``['001', '002']``.
    %(verbose)s

    Returns
    -------
    entity_vals : list of str
        List of the values associated with an `entity_key` in the BIDS dataset
        pointed to by `root`.

    Examples
    --------
    >>> root = Path('./mne_bids/tests/data/tiny_bids').absolute()
    >>> entity_key = 'subject'
    >>> get_entity_vals(root, entity_key)
    ['01']
    >>> get_entity_vals(root, entity_key, with_key=True)
    ['sub-01']

    Notes
    -----
    This function will scan the entire ``root``, except for a
    ``derivatives`` subfolder placed directly under ``root``.

    References
    ----------
    .. [1] https://bids-specification.rtfd.io/en/latest/common-principles.html#entities

    """
    root = _check_fname(
        fname=root,
        overwrite="read",
        must_exist=True,
        need_dir=True,
        name="Root directory",
    )
    root = Path(root).expanduser()

    entities = (
        "subject",
        "task",
        "session",
        "acquisition",
        "run",
        "processing",
        "space",
        "recording",
        "split",
        "description",
        "suffix",
    )
    entities_abbr = (
        "sub",
        "task",
        "ses",
        "acq",
        "run",
        "proc",
        "space",
        "rec",
        "split",
        "desc",
        "suffix",
    )
    entity_long_abbr_map = dict(zip(entities, entities_abbr))

    if entity_key not in entities:
        raise ValueError(
            f'`key` must be one of: {", ".join(entities)}. ' f"Got: {entity_key}"
        )

    ignore_subjects = _ensure_tuple(ignore_subjects)
    ignore_sessions = _ensure_tuple(ignore_sessions)
    ignore_tasks = _ensure_tuple(ignore_tasks)
    ignore_acquisitions = _ensure_tuple(ignore_acquisitions)
    ignore_runs = _ensure_tuple(ignore_runs)
    ignore_processings = _ensure_tuple(ignore_processings)
    ignore_spaces = _ensure_tuple(ignore_spaces)
    ignore_recordings = _ensure_tuple(ignore_recordings)
    ignore_splits = _ensure_tuple(ignore_splits)
    ignore_descriptions = _ensure_tuple(ignore_descriptions)
    ignore_modalities = _ensure_tuple(ignore_modalities)

    ignore_dirs = _ensure_tuple(ignore_dirs)
    existing_ignore_dirs = [
        root / d for d in ignore_dirs if (root / d).exists() and (root / d).is_dir()
    ]
    ignore_dirs = _ensure_tuple(existing_ignore_dirs)

    p = re.compile(rf"{entity_long_abbr_map[entity_key]}-(.*?)_")
    values = list()
    filenames = root.glob(f"**/*{entity_long_abbr_map[entity_key]}-*_*")

    for filename in filenames:
        # Skip ignored directories
        # XXX In Python 3.9, we can use Path.is_relative_to() here
        if any(
            [str(filename).startswith(str(ignore_dir)) for ignore_dir in ignore_dirs]
        ):
            continue

        if ignore_datatypes and filename.parent.name in ignore_datatypes:
            continue
        if ignore_subjects and any(
            [filename.stem.startswith(f"sub-{s}_") for s in ignore_subjects]
        ):
            continue
        if ignore_sessions and any(
            [f"_ses-{s}_" in filename.stem for s in ignore_sessions]
        ):
            continue
        if ignore_tasks and any([f"_task-{t}_" in filename.stem for t in ignore_tasks]):
            continue
        if ignore_acquisitions and any(
            [f"_acq-{a}_" in filename.stem for a in ignore_acquisitions]
        ):
            continue
        if ignore_runs and any([f"_run-{r}_" in filename.stem for r in ignore_runs]):
            continue
        if ignore_processings and any(
            [f"_proc-{p}_" in filename.stem for p in ignore_processings]
        ):
            continue
        if ignore_spaces and any(
            [f"_space-{s}_" in filename.stem for s in ignore_spaces]
        ):
            continue
        if ignore_recordings and any(
            [f"_rec-{a}_" in filename.stem for a in ignore_recordings]
        ):
            continue
        if ignore_splits and any(
            [f"_split-{s}_" in filename.stem for s in ignore_splits]
        ):
            continue
        if ignore_descriptions and any(
            [f"_desc-{d}_" in filename.stem for d in ignore_descriptions]
        ):
            continue
        if ignore_modalities and any(
            [f"_{k}" in filename.stem for k in ignore_modalities]
        ):
            continue

        match = p.search(filename.stem)
        value = match.group(1)
        if with_key:
            value = f"{entity_long_abbr_map[entity_key]}-{value}"
        if value not in values:
            values.append(value)
    return sorted(values)


def _mkdir_p(path, overwrite=False):
    """Create a directory, making parent directories as needed [1].

    References
    ----------
    .. [1] stackoverflow.com/questions/600268/mkdir-p-functionality-in-python

    """
    if overwrite and op.isdir(path):
        sh.rmtree(path)
        logger.info(f"Clearing path: {path}")

    os.makedirs(path, exist_ok=True)
    if not op.isdir(path):
        logger.info(f"Creating folder: {path}")


def _find_best_candidates(params, candidate_list):
    """Return the best candidate, based on the number of param matches.

    Assign each candidate a score, based on how many entities are shared with
    the ones supplied in the `params` parameter. The candidate with the highest
    score wins. Candidates with entities that conflict with the supplied
    `params` are disqualified.

    Parameters
    ----------
    params : dict
        The entities that the candidate should match.
    candidate_list : list of str
        A list of candidate filenames.

    Returns
    -------
    best_candidates : list of str
        A list of all the candidate filenames that are tied for first place.
        Hopefully, the list will have a length of one.
    """
    params = {key: value for key, value in params.items() if value is not None}

    best_candidates = []
    best_n_matches = 0
    for candidate in candidate_list:
        n_matches = 0
        candidate_disqualified = False
        candidate_params = get_entities_from_fname(candidate)
        for entity, value in params.items():
            if entity in candidate_params:
                if candidate_params[entity] is None:
                    continue
                elif candidate_params[entity] == value:
                    n_matches += 1
                else:
                    # Incompatible entity found, candidate is disqualified
                    candidate_disqualified = True
                    break
        if not candidate_disqualified:
            if n_matches > best_n_matches:
                best_n_matches = n_matches
                best_candidates = [candidate]
            elif n_matches == best_n_matches:
                best_candidates.append(candidate)
    return best_candidates


def _get_datatypes_for_sub(*, root, sub, ses=None):
    """Retrieve data modalities for a specific subject and session."""
    subject_dir = op.join(root, f"sub-{sub}")
    if ses is not None:
        subject_dir = op.join(subject_dir, f"ses-{ses}")

    # TODO We do this to ensure we don't accidentally pick up any "spurious"
    # TODO sub-directories. But is that really necessary with valid BIDS data?
    modalities_in_dataset = get_datatypes(root=root)
    subdirs = [f.name for f in os.scandir(subject_dir) if f.is_dir()]
    available_modalities = [s for s in subdirs if s in modalities_in_dataset]
    return available_modalities


def _infer_datatype(*, root, sub, ses):
    # Check which suffix is available for this particular
    # subject & session. If we get no or multiple hits, throw an error.

    modalities = _get_datatypes_for_sub(root=root, sub=sub, ses=ses)

    # We only want to handle electrophysiological data here.
    allowed_recording_modalities = ["meg", "eeg", "ieeg"]
    modalities = list(set(modalities) & set(allowed_recording_modalities))
    if not modalities:
        raise ValueError("No electrophysiological data found.")
    elif len(modalities) >= 2:
        msg = (
            f"Found data of more than one recording datatype. Please "
            f"pass the `suffix` parameter to specify which data to load. "
            f"Found the following modalitiess: {modalities}"
        )
        raise RuntimeError(msg)

    assert len(modalities) == 1
    return modalities[0]


def _path_to_str(var):
    """Make sure var is a string or Path, return string representation."""
    if not isinstance(var, (Path, str)):
        raise ValueError(
            f"All path parameters must be either strings or "
            f"pathlib.Path objects. Found type {type(var)}."
        )
    else:
        return str(var)


def _filter_fnames(
    fnames,
    *,
    subject=None,
    session=None,
    task=None,
    acquisition=None,
    run=None,
    processing=None,
    recording=None,
    space=None,
    split=None,
    description=None,
    suffix=None,
    extension=None,
):
    """Filter a list of BIDS filenames / paths based on BIDS entity values.

    Input can be str or list of str.

    Parameters
    ----------
    fnames : iterable of pathlib.Path | iterable of str

    Returns
    -------
    list of pathlib.Path

    """
    subject = _ensure_tuple(subject)
    session = _ensure_tuple(session)
    task = _ensure_tuple(task)
    acquisition = _ensure_tuple(acquisition)
    run = _ensure_tuple(run)
    processing = _ensure_tuple(processing)
    space = _ensure_tuple(space)
    recording = _ensure_tuple(recording)
    split = _ensure_tuple(split)
    description = _ensure_tuple(description)
    suffix = _ensure_tuple(suffix)
    extension = _ensure_tuple(extension)

    leading_path_str = r".*\/?"  # nothing or something ending with a `/`
    sub_str = r"sub-(" + "|".join(subject) + ")" if subject else r"sub-([^_]+)"
    ses_str = r"_ses-(" + "|".join(session) + ")" if session else r"(|_ses-([^_]+))"
    task_str = r"_task-(" + "|".join(task) + ")" if task else r"(|_task-([^_]+))"
    acq_str = (
        r"_acq-(" + "|".join(acquisition) + ")" if acquisition else r"(|_acq-([^_]+))"
    )
    run_str = r"_run-(" + "|".join(run) + ")" if run else r"(|_run-([^_]+))"
    proc_str = (
        r"_proc-(" + "|".join(processing) + ")" if processing else r"(|_proc-([^_]+))"
    )
    space_str = r"_space-(" + "|".join(space) + ")" if space else r"(|_space-([^_]+))"
    rec_str = r"_rec-(" + "|".join(recording) + ")" if recording else r"(|_rec-([^_]+))"
    split_str = r"_split-(" + "|".join(split) + ")" if split else r"(|_split-([^_]+))"
    desc_str = (
        r"_desc-(" + "|".join(description) + ")" if description else r"(|_desc-([^_]+))"
    )
    suffix_str = r"_(" + "|".join(suffix) + ")" if suffix else r"_([^_]+)"
    ext_str = r"(" + "|".join(extension) + ")" if extension else r".([^_]+)"

    regexp = (
        leading_path_str
        + sub_str
        + ses_str
        + task_str
        + acq_str
        + run_str
        + proc_str
        + space_str
        + rec_str
        + split_str
        + desc_str
        + suffix_str
        + ext_str
    )

    # Convert to str so we can apply the regexp ...
    fnames = [str(f) for f in fnames]

    # https://stackoverflow.com/a/51246151/1944216
    fnames_filtered = sorted(filter(re.compile(regexp).match, fnames))

    # ... and return Paths.
    fnames_filtered = [Path(f) for f in fnames_filtered]
    return fnames_filtered


def find_matching_paths(
    root,
    subjects=None,
    sessions=None,
    tasks=None,
    acquisitions=None,
    runs=None,
    processings=None,
    recordings=None,
    spaces=None,
    splits=None,
    descriptions=None,
    suffixes=None,
    extensions=None,
    datatypes=None,
    check=False,
):
    """Get list of all matching paths for all matching entity values.

    Input can be str or list of str. None matches all found values.

    Performs a recursive search, starting in ``.root`` (if set), based on
    `BIDSPath.entities` object.

    Parameters
    ----------
    root : pathlib.Path | str
        The root of the BIDS path.
    subjects : str | array-like of str | None
        The subject ID. Corresponds to "sub".
    sessions : str | array-like of str | None
        The acquisition session. Corresponds to "ses".
    tasks : str | array-like of str | None
        The experimental task. Corresponds to "task".
    acquisitions: str | array-like of str | None
        The acquisition parameters. Corresponds to "acq".
    runs : str | array-like of str | None
        The run number. Corresponds to "run".
    processings : str | array-like of str | None
        The processing label. Corresponds to "proc".
    recordings : str | array-like of str | None
        The recording name. Corresponds to "rec".
    spaces : str | array-like of str | None
        The coordinate space for anatomical and sensor location
        files (e.g., ``*_electrodes.tsv``, ``*_markers.mrk``).
        Corresponds to "space".
        Note that valid values for ``space`` must come from a list
        of BIDS keywords as described in the BIDS specification.
    splits : str | array-like of str | None
        The split of the continuous recording file for ``.fif`` data.
        Corresponds to "split".
    descriptions : str | array-like of str | None
        This corresponds to the BIDS entity ``desc``. It is used to provide
        additional information for derivative data, e.g., preprocessed data
        may be assigned ``description='cleaned'``.

        .. versionadded:: 0.11
    suffixes : str | array-like of str | None
        The filename suffix. This is the entity after the
        last ``_`` before the extension. E.g., ``'channels'``.
        The following filename suffix's are accepted:
        'meg', 'markers', 'eeg', 'ieeg', 'T1w',
        'participants', 'scans', 'electrodes', 'coordsystem',
        'channels', 'events', 'headshape', 'digitizer',
        'beh', 'physio', 'stim'
    extensions : str | array-like of str | None
        The extension of the filename. E.g., ``'.json'``.
    datatypes : str | array-like of str | None
        The BIDS data type, e.g., ``'anat'``, ``'func'``, ``'eeg'``, ``'meg'``,
        ``'ieeg'``.
    check : bool
        If ``True``, only returns paths that conform to BIDS. If ``False``
        (default), the ``.check`` attribute of the returned
        `mne_bids.BIDSPath` object will be set to ``True`` for paths that
        do conform to BIDS, and to ``False`` for those that don't.

    Returns
    -------
    bids_paths : list of mne_bids.BIDSPath
        The matching paths.

    """
    fpaths = _return_root_paths(root, datatype=datatypes, ignore_json=False)

    fpaths_filtered = _filter_fnames(
        fpaths,
        subject=subjects,
        session=sessions,
        task=tasks,
        acquisition=acquisitions,
        run=runs,
        processing=processings,
        recording=recordings,
        space=spaces,
        split=splits,
        description=descriptions,
        suffix=suffixes,
        extension=extensions,
    )

    bids_paths = _fnames_to_bidspaths(fpaths_filtered, root, check=check)
    return bids_paths


def _return_root_paths(root, datatype=None, ignore_json=True):
    """Return all paths in root.

    Can be filtered by datatype (which is present in the path but not in
    the BIDSPath basename). Can also be list of datatypes.
    root : pathlib.Path | str
        The root of the BIDS path.
    datatype : str | array-like of str | None
        The BIDS data type, e.g., ``'anat'``, ``'func'``, ``'eeg'``, ``'meg'``,
        ``'ieeg'``.
    """
    root = Path(root)  # if root is str

    if datatype is not None:
        datatype = _ensure_tuple(datatype)
        search_str = f'*/{"|".join(datatype)}/*'
    else:
        search_str = "*.*"

    paths = root.rglob(search_str)
    # Only keep files (not directories), and omit the JSON sidecars
    # if ignore_json is True.
    if ignore_json:
        paths = [p for p in paths if p.is_file() and p.suffix != ".json"]
    else:
        paths = [p for p in paths if p.is_file()]

    return paths


def _fnames_to_bidspaths(fnames, root, check=False):
    """Make BIDSPaths from file names.

    To check whether the BIDSPath is conforming to BIDS if check=True, we
    first instantiate without checking and then run the check manually,
    allowing us to be more specific about the exception to catch.

    Parameters
    ----------
    fnames : list of str
        Filenames as list of strings.
    root : path-like | None
        The root directory of the BIDS dataset.
    check : bool
        If ``True``, only returns paths that conform to BIDS. If ``False``
        (default), the ``.check`` attribute of the returned
        `mne_bids.BIDSPath` object will be set to ``True`` for paths that
        do conform to BIDS, and to ``False`` for those that don't.

    Returns
    -------
    bids_paths : list of mne_bids.BIDSPath
        Bids paths.

    """
    bids_paths = []
    for fname in fnames:
        datatype = _infer_datatype_from_path(fname)
        bids_path = get_bids_path_from_fname(fname, check=False)
        bids_path.root = root
        bids_path.datatype = datatype
        bids_path.check = True

        try:
            bids_path._check()
        except ValueError:
            # path is not BIDS-compatible
            if check:  # skip!
                continue
            else:
                bids_path.check = False

        bids_paths.append(bids_path)
    return bids_paths



r"""Write Elekta/Neuromag/MEGIN fine-calibration data to BIDS.

example usage:
$ mne_bids calibration_to_bids --subject_id=01 --session=test
--bids_root=bids_root --file=sss_cal.dat

"""
# Authors: Richard Höchenberger <richard.hoechenberger@gmail.com>
#
# License: BSD-3-Clause

from mne.utils import logger

import mne_bids
from mne_bids import BIDSPath, write_meg_calibration


def run():
    """Run the calibration_to_bids command."""
    from mne.commands.utils import get_optparser

    parser = get_optparser(
        __file__,
        usage="usage: %prog options args",
        prog_prefix="mne_bids",
        version=mne_bids.__version__,
    )

    parser.add_option(
        "--bids_root",
        dest="bids_root",
        help="The path of the folder containing the BIDS dataset",
    )
    parser.add_option("--subject_id", dest="subject", help="Subject name")
    parser.add_option("--session_id", dest="session", help="Session name")
    parser.add_option("--file", dest="fname", help="The path of the crosstalk file")
    parser.add_option(
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Whether do generate additional diagnostic output",
    )

    opt, args = parser.parse_args()
    if args:
        parser.print_help()
        parser.error(
            f"Please do not specify arguments without flags. " f"Got: {args}.\n"
        )

    if opt.bids_root is None:
        parser.print_help()
        parser.error("You must specify bids_root")
    if opt.subject is None:
        parser.print_help()
        parser.error("You must specify a subject")

    bids_path = BIDSPath(subject=opt.subject, session=opt.session, root=opt.bids_root)

    logger.info(f"Writing fine-calibration file {bids_path.basename} …")
    write_meg_calibration(
        calibration=opt.fname, bids_path=bids_path, verbose=opt.verbose
    )


if __name__ == "__main__":
    run()



"""Define coil types for MEG."""

# Authors: Matt Sanderson <matt.sanderson@mq.edu.au>
#
# License: BSD-3-Clause
from mne.io.constants import FIFF


def get_coil_types():
    """Return all known coil types.

    Returns
    -------
    coil_types : dict
        The keys contain the channel types, and the values contain the
        corresponding values in the info['chs'][idx]['kind']

    """
    return dict(
        meggradaxial=(
            FIFF.FIFFV_COIL_KIT_GRAD,
            FIFF.FIFFV_COIL_CTF_GRAD,
            # Support for gradient-compensated data:
            int(FIFF.FIFFV_COIL_CTF_GRAD | (3 << 16)),
            int(FIFF.FIFFV_COIL_CTF_GRAD | (2 << 16)),
            FIFF.FIFFV_COIL_AXIAL_GRAD_5CM,
            FIFF.FIFFV_COIL_BABY_GRAD,
        ),
        megrefgradaxial=(
            FIFF.FIFFV_COIL_CTF_REF_GRAD,
            FIFF.FIFFV_COIL_CTF_OFFDIAG_REF_GRAD,
            FIFF.FIFFV_COIL_MAGNES_REF_GRAD,
            FIFF.FIFFV_COIL_MAGNES_OFFDIAG_REF_GRAD,
        ),
        meggradplanar=(
            FIFF.FIFFV_COIL_VV_PLANAR_T1,
            FIFF.FIFFV_COIL_VV_PLANAR_T2,
            FIFF.FIFFV_COIL_VV_PLANAR_T3,
        ),
        megmag=(
            FIFF.FIFFV_COIL_POINT_MAGNETOMETER,
            FIFF.FIFFV_COIL_VV_MAG_W,
            FIFF.FIFFV_COIL_VV_MAG_T1,
            FIFF.FIFFV_COIL_VV_MAG_T2,
            FIFF.FIFFV_COIL_VV_MAG_T3,
            FIFF.FIFFV_COIL_NM_122,
            FIFF.FIFFV_COIL_MAGNES_MAG,
            FIFF.FIFFV_COIL_BABY_MAG,
        ),
        megrefmag=(
            FIFF.FIFFV_COIL_KIT_REF_MAG,
            FIFF.FIFFV_COIL_CTF_REF_MAG,
            FIFF.FIFFV_COIL_MAGNES_REF_MAG,
            FIFF.FIFFV_COIL_BABY_REF_MAG,
            FIFF.FIFFV_COIL_BABY_REF_MAG2,
            FIFF.FIFFV_COIL_ARTEMIS123_REF_MAG,
            FIFF.FIFFV_COIL_MAGNES_REF_MAG,
        ),
        eeg=(FIFF.FIFFV_COIL_EEG,),
        misc=(FIFF.FIFFV_COIL_NONE,),
    )


def coil_type(info, idx, ch_type="n/a"):
    """Get coil type.

    Parameters
    ----------
    info : dict
        Measurement info
    idx : int
        Index of channel
    ch_type : str
        Channel type to fall back upon if a more specific
        type is not found

    Returns
    -------
    type : 'meggradaxial' | 'megrefgradaxial' | 'meggradplanar'
           'megmag' | 'megrefmag' | 'eeg' | 'misc'
        Type of coil

    """
    ch = info["chs"][idx]
    for key, values in get_coil_types().items():
        if ch["coil_type"] in values:
            return key
    return ch_type



"""Make BIDS report from dataset and sidecar files."""

# Authors: Adam Li <adam2392@gmail.com>
#
# License: BSD-3-Clause
import json
import os.path as op
import textwrap
from pathlib import Path

import jinja2
import numpy as np
from mne.utils import logger, verbose

from mne_bids.config import ALLOWED_DATATYPES, DOI
from mne_bids.path import (
    BIDSPath,
    _find_matching_sidecar,
    _parse_ext,
    get_bids_path_from_fname,
    get_datatypes,
    get_entity_vals,
)
from mne_bids.tsv_handler import _from_tsv
from mne_bids.utils import warn

jinja_env = jinja2.Environment(
    loader=jinja2.PackageLoader(
        package_name="mne_bids.report", package_path="templates"
    )
)


def _pretty_str(listed):
    # make strings a sequence of ',' and 'and'
    if not isinstance(listed, list):
        listed = list(listed)

    if len(listed) <= 1:
        return ",".join(listed)
    return "{}, and {}".format(", ".join(listed[:-1]), listed[-1])


def _range_str(minval, maxval, meanval, stdval, n_unknown, type):
    if minval == "n/a":
        return "ages all unknown"

    if n_unknown > 0:
        unknown_str = f"; {n_unknown} with unknown {type}"
    else:
        unknown_str = ""
    return (
        f"ages ranged from {round(minval, 2)} to {round(maxval, 2)} "
        f"(mean = {round(meanval, 2)}, std = {round(stdval, 2)}{unknown_str})"
    )


def _summarize_participant_hand(hands):
    n_unknown = len([hand for hand in hands if hand == "n/a"])

    if n_unknown == len(hands):
        return "handedness were all unknown"

    n_rhand = len([hand for hand in hands if hand.upper() == "R"])
    n_lhand = len([hand for hand in hands if hand.upper() == "L"])
    n_ambidex = len([hand for hand in hands if hand.upper() == "A"])

    return (
        f"comprised of {n_rhand} right hand, {n_lhand} left hand "
        f"and {n_ambidex} ambidextrous"
    )


def _summarize_participant_sex(sexs):
    n_unknown = len([sex for sex in sexs if sex == "n/a"])

    if n_unknown == len(sexs):
        return "sex were all unknown"

    n_males = len([sex for sex in sexs if sex.upper() == "M"])
    n_females = len([sex for sex in sexs if sex.upper() == "F"])

    return f"comprised of {n_males} male and {n_females} female participants"


def _length_recording_str(length_recordings):
    import numpy as np

    if length_recordings is None:
        return ""

    min_record_length = round(np.min(length_recordings), 2)
    max_record_length = round(np.max(length_recordings), 2)
    mean_record_length = round(np.mean(length_recordings), 2)
    std_record_length = round(np.std(length_recordings), 2)
    total_record_length = round(sum(length_recordings), 2)

    return (
        f"Recording durations ranged from {min_record_length} to "
        f"{max_record_length} seconds "
        f"(mean = {mean_record_length}, std = {std_record_length}), "
        f"for a total of {total_record_length} seconds of data recorded "
        f"over all scans."
    )


def _summarize_software_filters(software_filters):
    if software_filters in [{}, "n/a"]:
        return ""

    msg = ""
    for key, value in software_filters.items():
        msg += f"{key}"

        if isinstance(value, dict) and value:
            parameters = []
            for param_name, param_value in value.items():
                if param_name and param_value:
                    parameters.append(f"{param_value} {param_name}")
            if parameters:
                msg += " with parameters "
                msg += ", ".join(parameters)
    return msg


def _pretty_dict(template_dict):
    """Remove problematic blank spaces."""
    for key, val in template_dict.items():
        if val == " ":
            template_dict[key] = "n/a"


def _summarize_dataset(root):
    """Summarize the dataset_desecription.json file.

    Required dataset descriptors include:
        - Name
        - BIDSVersion

    Added descriptors include:
        - Authors
        - DOI

    Parameters
    ----------
    root : path-like
        The path of the root of the BIDS compatible folder.

    Returns
    -------
    template_dict : dict
        A dictionary of values for various template strings.
    """
    dataset_descrip_fpath = op.join(root, "dataset_description.json")
    if not op.exists(dataset_descrip_fpath):
        return dict()

    # read file and 'REQUIRED' components of it
    with open(dataset_descrip_fpath, encoding="utf-8-sig") as fin:
        dataset_description = json.load(fin)

    # create dictionary to pass into template string
    name = dataset_description["Name"]
    bids_version = dataset_description["BIDSVersion"]
    authors = dataset_description["Authors"]
    template_dict = {
        "name": name,
        "bids_version": bids_version,
        "mne_bids_doi": DOI,
        "authors": _pretty_str(authors),
    }
    _pretty_dict(template_dict)
    return template_dict


def _summarize_participants_tsv(root):
    """Summarize `participants.tsv` file in BIDS root directory.

    Parameters
    ----------
    root : path-like
        The path of the root of the BIDS compatible folder.

    Returns
    -------
    template_dict : dict
        A dictionary of values for various template strings.
    """
    participants_tsv_fpath = op.join(root, "participants.tsv")
    if not op.exists(participants_tsv_fpath):
        return dict()

    participants_tsv = _from_tsv(str(participants_tsv_fpath))
    p_ids = participants_tsv["participant_id"]
    logger.info(f"Summarizing participants.tsv {participants_tsv_fpath}...")

    # summarize sex count statistics
    keys = ["M", "F", "n/a"]
    p_sex = participants_tsv.get("sex")
    # phrasing works for both sex and gender
    p_gender = participants_tsv.get("gender")
    sexs = ["n/a"]
    if p_sex or p_gender:
        # only summarize sex if it conforms to `keys` referenced above
        p_sex = p_gender if p_sex is None else p_sex
        if all([sex.upper() in keys for sex in p_sex if sex != "n/a"]):
            sexs = p_sex

    # summarize hand count statistics
    keys = ["R", "L", "A", "n/a"]
    p_hands = participants_tsv.get("hand")
    hands = ["n/a"]
    if p_hands:
        # only summarize handedness if it conforms to
        # mne-bids handedness
        if all([hand.upper() in keys for hand in p_hands if hand != "n/a"]):
            hands = p_hands

    # summarize age statistics: mean, std, min, max
    p_ages = participants_tsv.get("age")
    min_age, max_age = "n/a", "n/a"
    mean_age, std_age = "n/a", "n/a"
    n_age_unknown = len(p_ages) if p_ages else len(p_ids)
    if p_ages:
        # only summarize age if they are numerics
        if all([age.isnumeric() for age in p_ages if age != "n/a"]):
            age_list = [float(age) for age in p_ages if age != "n/a"]
            n_age_unknown = len(p_ids) - len(age_list)
            if age_list:
                min_age, max_age = np.min(age_list), np.max(age_list)
                mean_age, std_age = np.mean(age_list), np.std(age_list)

    template_dict = {
        "sexs": _summarize_participant_sex(sexs),
        "hands": _summarize_participant_hand(hands),
        "ages": _range_str(min_age, max_age, mean_age, std_age, n_age_unknown, "age"),
    }
    return template_dict


def _summarize_scans(root, session=None):
    """Summarize scans in BIDS root directory.

    Summarizes scans only if there is a *_scans.tsv file.

    Parameters
    ----------
    root : path-like
        The path of the root of the BIDS compatible folder.
    session : str, optional
        The session for a item. Corresponds to "ses".

    Returns
    -------
    template_dict : dict
        A dictionary of values for various template strings.

    """
    root = Path(root)
    if session is None:
        search_str = "*_scans.tsv"
    else:
        search_str = f"*ses-{session}" f"*_scans.tsv"
    scans_fpaths = list(root.rglob(search_str))
    if len(scans_fpaths) == 0:
        warn(
            "No *scans.tsv files found. Currently, "
            "we do not generate a report without the scans.tsv files."
        )
        return dict()

    logger.info(f"Summarizing scans.tsv files {scans_fpaths}...")

    # summarize sidecar.json, channels.tsv template
    sidecar_dict = _summarize_sidecar_json(root, scans_fpaths)
    channels_dict = _summarize_channels_tsv(root, scans_fpaths)
    template_dict = dict()
    template_dict.update(**sidecar_dict)
    template_dict.update(**channels_dict)

    return template_dict


def _summarize_sidecar_json(root, scans_fpaths):
    """Summarize scans in BIDS root directory.

    Parameters
    ----------
    root : path-like
        The path of the root of the BIDS compatible folder.
    scans_fpaths : list
        A list of all *_scans.tsv files in ``root``. The summary
        will occur for all scans listed in the *_scans.tsv files.

    Returns
    -------
    template_dict : dict
        A dictionary of values for various template strings.

    """
    n_scans = 0
    powerlinefreqs, sfreqs = set(), set()
    manufacturers = set()
    length_recordings = []

    # loop through each scan
    for scan_fpath in scans_fpaths:
        # load in the scans.tsv file
        # and read metadata for each scan
        scans_tsv = _from_tsv(scan_fpath)
        scans = scans_tsv["filename"]
        for scan in scans:
            # summarize metadata of recordings
            bids_path, ext = _parse_ext(scan)
            datatype = op.dirname(scan)
            if datatype not in ALLOWED_DATATYPES:
                continue

            n_scans += 1

            # convert to BIDSPath
            if not isinstance(bids_path, BIDSPath):
                bids_path = get_bids_path_from_fname(bids_path)
            bids_path.root = root

            # XXX: improve to allow emptyroom
            if bids_path.subject == "emptyroom":
                continue

            sidecar_fname = _find_matching_sidecar(
                bids_path=bids_path, suffix=datatype, extension=".json"
            )
            with open(sidecar_fname, encoding="utf-8-sig") as fin:
                sidecar_json = json.load(fin)

            # aggregate metadata from each scan
            # REQUIRED kwargs
            sfreq = sidecar_json["SamplingFrequency"]
            powerlinefreq = str(sidecar_json["PowerLineFrequency"])
            software_filters = sidecar_json.get("SoftwareFilters")
            if not software_filters:
                software_filters = "n/a"

            # RECOMMENDED kwargs
            manufacturer = sidecar_json.get("Manufacturer", "n/a")
            record_duration = sidecar_json.get("RecordingDuration", "n/a")

            sfreqs.add(str(np.round(sfreq, 2)))
            powerlinefreqs.add(str(powerlinefreq))
            if manufacturer != "n/a":
                manufacturers.add(manufacturer)
            length_recordings.append(record_duration)

    # XXX: length summary is only allowed, if no 'n/a' was found
    if any([dur == "n/a" for dur in length_recordings]):
        length_recordings = None

    template_dict = {
        "n_scans": n_scans,
        "manufacturer": _pretty_str(manufacturers),
        "sfreq": _pretty_str(sfreqs),
        "powerlinefreq": _pretty_str(powerlinefreqs),
        "software_filters": _summarize_software_filters(software_filters),
        "length_recordings": _length_recording_str(length_recordings),
    }
    return template_dict


def _summarize_channels_tsv(root, scans_fpaths):
    """Summarize channels.tsv data in BIDS root directory.

    Currently, summarizes all REQUIRED components of channels
    data, and some RECOMMENDED and OPTIONAL components.

    Parameters
    ----------
    root : path-like
        The path of the root of the BIDS compatible folder.
    scans_fpaths : list
        A list of all *_scans.tsv files in ``root``. The summary
        will occur for all scans listed in the *_scans.tsv files.

    Returns
    -------
    template_dict : dict
        A dictionary of values for various template strings.
    """
    root = Path(root)

    # keep track of channel type, status
    ch_status_count = {"bad": [], "good": []}
    ch_count = []

    # loop through each scan
    for scan_fpath in scans_fpaths:
        # load in the scans.tsv file
        # and read metadata for each scan
        scans_tsv = _from_tsv(scan_fpath)
        scans = scans_tsv["filename"]
        for scan in scans:
            # summarize metadata of recordings
            bids_path, _ = _parse_ext(scan)
            datatype = op.dirname(scan)
            if datatype not in ["meg", "eeg", "ieeg"]:
                continue

            # convert to BIDSPath
            if not isinstance(bids_path, BIDSPath):
                bids_path = get_bids_path_from_fname(bids_path)
            bids_path.root = root

            # XXX: improve to allow emptyroom
            if bids_path.subject == "emptyroom":
                continue

            channels_fname = _find_matching_sidecar(
                bids_path=bids_path, suffix="channels", extension=".tsv"
            )

            # summarize channels.tsv
            channels_tsv = _from_tsv(channels_fname)
            for status in ch_status_count.keys():
                ch_status = [ch for ch in channels_tsv["status"] if ch == status]
                ch_status_count[status].append(len(ch_status))
            ch_count.append(len(channels_tsv["name"]))

    # create summary template strings for status
    template_dict = {
        "mean_chs": np.mean(ch_count),
        "std_chs": np.std(ch_count),
        "mean_good_chs": np.mean(ch_status_count["good"]),
        "std_good_chs": np.std(ch_status_count["good"]),
        "mean_bad_chs": np.mean(ch_status_count["bad"]),
        "std_bad_chs": np.std(ch_status_count["bad"]),
    }
    for key, val in template_dict.items():
        template_dict[key] = round(val, 2)
    return template_dict


@verbose
def make_report(root, session=None, verbose=None):
    """Create a methods paragraph string from BIDS dataset.

    Summarizes the REQUIRED components in the BIDS specification
    and also some RECOMMENDED components. Currently, the methods
    paragraph summarize the:

      - dataset_description.json file
      - (optional) participants.tsv file
      - (optional) datatype-agnostic files for (M/I)EEG data,
        which reads files from the ``*_scans.tsv`` file.

    Parameters
    ----------
    root : path-like
        The path of the root of the BIDS compatible folder.
    session : str | None
            The (optional) session for a item. Corresponds to "ses".
    %(verbose)s

    Returns
    -------
    paragraph : str
        The paragraph wrapped with 80 characters per line
        describing the summary of the subjects.
    """
    # high level summary
    subjects = get_entity_vals(root, entity_key="subject")
    sessions = get_entity_vals(root, entity_key="session")
    modalities = get_datatypes(root)

    # only summarize allowed modalities (MEG/EEG/iEEG) data
    # map them to a pretty looking string
    datatype_map = {
        "meg": "MEG",
        "eeg": "EEG",
        "ieeg": "iEEG",
    }
    modalities = [
        datatype_map[datatype]
        for datatype in modalities
        if datatype in datatype_map.keys()
    ]

    # REQUIRED: dataset_description.json summary
    dataset_summary = _summarize_dataset(root)

    # RECOMMENDED: participants summary
    participant_summary = _summarize_participants_tsv(root)

    # RECOMMENDED: scans summary
    scans_summary = _summarize_scans(root, session=session)

    dataset_agnostic_summary = scans_summary.copy()
    dataset_agnostic_summary["system"] = _pretty_str(modalities)

    # turn off 'recommended' report summary
    # if files are not available to summarize
    if not participant_summary:
        participants_info = ""
    else:
        particpants_info_template = jinja_env.get_template("participants.jinja")
        participants_info = particpants_info_template.render(**participant_summary)
        logger.info(f"The participant template found: {participants_info}")

    if not scans_summary:
        datatype_agnostic_info = ""
    else:
        datatype_agnostic_template = jinja_env.get_template("datatype_agnostic.jinja")
        datatype_agnostic_info = datatype_agnostic_template.render(
            **dataset_agnostic_summary
        )

    dataset_summary.update(
        {
            "n_subjects": len(subjects),
            "participants_info": participants_info,
            "n_sessions": len(sessions),
            "sessions": _pretty_str(sessions),
        }
    )

    # XXX: add channel summary for modalities (ieeg, meg, eeg)
    # create the content and mne Template
    # lower-case templates are "Recommended",
    # while upper-case templates are "Required".

    dataset_summary_template = jinja_env.get_template("dataset_summary.jinja")
    dataset_summary_info = dataset_summary_template.render(**dataset_summary)

    # Concatenate info and clean the paragraph
    paragraph = f"{dataset_summary_info}\n{datatype_agnostic_info}"
    paragraph = paragraph.replace("\n", " ")
    while "  " in paragraph:
        paragraph = paragraph.replace("  ", " ")

    return "\n".join(textwrap.wrap(paragraph, width=80))



"""Check whether a file format is supported by BIDS and then load it."""

# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#          Teon Brooks <teon.brooks@gmail.com>
#          Chris Holdgraf <choldgraf@berkeley.edu>
#          Stefan Appelhoff <stefan.appelhoff@mailbox.org>
#
# License: BSD-3-Clause
import json
import os
import os.path as op
import re
from datetime import datetime, timezone
from difflib import get_close_matches
from pathlib import Path

import mne
import numpy as np
from mne import events_from_annotations, io, pick_channels_regexp, read_events
from mne.coreg import fit_matched_points
from mne.transforms import apply_trans
from mne.utils import get_subjects_dir, logger

from mne_bids.config import (
    ALLOWED_DATATYPE_EXTENSIONS,
    ANNOTATIONS_TO_KEEP,
    _map_options,
    reader,
)
from mne_bids.dig import _read_dig_bids
from mne_bids.path import (
    BIDSPath,
    _find_matching_sidecar,
    _infer_datatype,
    _parse_ext,
    get_bids_path_from_fname,
)
from mne_bids.tsv_handler import _drop, _from_tsv
from mne_bids.utils import _get_ch_type_mapping, _import_nibabel, verbose, warn


def _read_raw(
    raw_path,
    electrode=None,
    hsp=None,
    hpi=None,
    allow_maxshield=False,
    config_path=None,
    **kwargs,
):
    """Read a raw file into MNE, making inferences based on extension."""
    _, ext = _parse_ext(raw_path)

    # KIT systems
    if ext in [".con", ".sqd"]:
        raw = io.read_raw_kit(
            raw_path, elp=electrode, hsp=hsp, mrk=hpi, preload=False, **kwargs
        )

    # BTi systems
    elif ext == ".pdf":
        raw = io.read_raw_bti(
            pdf_fname=raw_path,
            config_fname=config_path,
            head_shape_fname=hsp,
            preload=False,
            **kwargs,
        )

    elif ext == ".fif":
        raw = reader[ext](raw_path, allow_maxshield, **kwargs)

    elif ext in [".ds", ".vhdr", ".set", ".edf", ".bdf", ".EDF", ".snirf", ".cdt"]:
        raw_path = Path(raw_path)
        raw = reader[ext](raw_path, **kwargs)

    # MEF and NWB are allowed, but not yet implemented
    elif ext in [".mef", ".nwb"]:
        raise ValueError(
            f'Got "{ext}" as extension. This is an allowed '
            f"extension but there is no IO support for this "
            f"file format yet."
        )

    # No supported data found ...
    # ---------------------------
    else:
        raise ValueError(
            f"Raw file name extension must be one "
            f"of {ALLOWED_DATATYPE_EXTENSIONS}\n"
            f"Got {ext}"
        )
    return raw


def _read_events(events, event_id, raw, bids_path=None):
    """Retrieve events (for use in *_events.tsv) from FIFF/array & Annotations.

    Parameters
    ----------
    events : path-like | np.ndarray | None
        If a string, a path to an events file. If an array, an MNE events array
        (shape n_events, 3). If None, events will be generated from
        ``raw.annotations``.
    event_id : dict | None
        The event id dict used to create a 'trial_type' column in events.tsv,
        mapping a description key to an integer-valued event code.
    raw : mne.io.Raw
        The data as MNE-Python Raw object.
    bids_path : BIDSPath | None
        Can be used to determine if the data is a resting-state or empty-room
        recording, and will suppress a warning about missing events in this
        case.

    Returns
    -------
    all_events : np.ndarray, shape = (n_events, 3)
        The first column contains the event time in samples and the third
        column contains the event id. The second column is ignored for now but
        typically contains the value of the trigger channel either immediately
        before the event or immediately after.
    all_dur : np.ndarray, shape (n_events,)
        The event durations in seconds.
    all_desc : dict
        A dictionary with the keys corresponding to the event descriptions and
        the values to the event IDs.

    """
    # retrieve events
    if isinstance(events, np.ndarray):
        if events.ndim != 2:
            raise ValueError("Events must have two dimensions, " f"found {events.ndim}")
        if events.shape[1] != 3:
            raise ValueError(
                "Events must have second dimension of length 3, "
                f"found {events.shape[1]}"
            )
        events = events
    elif events is None:
        events = np.empty(shape=(0, 3), dtype=int)
    else:
        events = read_events(events).astype(int)

    if raw.annotations:
        if event_id is None:
            logger.info(
                "The provided raw data contains annotations, but you did not "
                'pass an "event_id" mapping from annotation descriptions to '
                "event codes. We will generate arbitrary event codes. "
                'To specify custom event codes, please pass "event_id".'
            )
        else:
            special_annots = {"BAD_ACQ_SKIP"}
            desc_without_id = sorted(
                set(raw.annotations.description) - set(event_id.keys())
            )
            # auto-add entries to `event_id` for "special" annotation values
            # (but only if they're needed)
            if set(desc_without_id) & special_annots:
                for annot in special_annots:
                    # use a value guaranteed to not be in use
                    event_id = {annot: max(event_id.values()) + 90000} | event_id
                # remove the "special" annots from the list of problematic annots
                desc_without_id = sorted(set(desc_without_id) - special_annots)
            if desc_without_id:
                raise ValueError(
                    f"The provided raw data contains annotations, but "
                    f'"event_id" does not contain entries for all annotation '
                    f"descriptions. The following entries are missing: "
                    f'{", ".join(desc_without_id)}'
                )

    # If we have events, convert them to Annotations so they can be easily
    # merged with existing Annotations.
    if events.size > 0:
        ids_without_desc = set(events[:, 2]) - set(event_id.values())
        if ids_without_desc:
            raise ValueError(
                f"No description was specified for the following event(s): "
                f'{", ".join([str(x) for x in sorted(ids_without_desc)])}. '
                f"Please add them to the event_id dictionary, or drop them "
                f"from the events array."
            )

        # Append events to raw.annotations. All event onsets are relative to
        # measurement beginning.
        id_to_desc_map = dict(zip(event_id.values(), event_id.keys()))
        # We don't pass `first_samp`, as set_annotations() below will take
        # care of this shift automatically.
        new_annotations = mne.annotations_from_events(
            events=events,
            sfreq=raw.info["sfreq"],
            event_desc=id_to_desc_map,
            orig_time=raw.annotations.orig_time,
        )

        raw = raw.copy()  # Don't alter the original.
        annotations = raw.annotations.copy()

        # We use `+=` here because `Annotations.__iadd__()` does the right
        # thing and also performs a sanity check on `Annotations.orig_time`.
        annotations += new_annotations
        raw.set_annotations(annotations)
        del id_to_desc_map, annotations, new_annotations

    # Now convert the Annotations to events.
    all_events, all_desc = events_from_annotations(
        raw,
        event_id=event_id,
        regexp=None,  # Include `BAD_` and `EDGE_` Annotations, too.
    )
    all_dur = raw.annotations.duration

    # Warn about missing events if not rest or empty-room data
    if (all_events.size == 0 and bids_path.task is not None) and (
        not bids_path.task.startswith("rest")
        or not (bids_path.subject == "emptyroom" and bids_path.task == "noise")
    ):
        warn(
            "No events found or provided. Please add annotations to the raw "
            "data, or provide the events and event_id parameters. For "
            "resting state data, BIDS recommends naming the task using "
            'labels beginning with "rest".'
        )

    return all_events, all_dur, all_desc


def _verbose_list_index(lst, val, *, allow_all=False):
    # try to "return lst.index(val)" for list of str, but be more
    # informative/verbose when it fails
    try:
        return lst.index(val)
    except ValueError as exc:
        # Use str cast here to deal with pathlib.Path instances
        extra = get_close_matches(str(val), [str(ll) for ll in lst])
        if allow_all and not extra:
            extra = lst
        extra = f". Did you mean one of {extra}?" if extra else ""
        raise ValueError(f"{exc}{extra}") from None


def _handle_participants_reading(participants_fname, raw, subject):
    participants_tsv = _from_tsv(participants_fname)
    subjects = participants_tsv["participant_id"]
    row_ind = _verbose_list_index(subjects, subject, allow_all=True)
    raw.info["subject_info"] = dict()  # start from scratch

    # set data from participants tsv into subject_info
    for col_name, value in participants_tsv.items():
        if col_name in ("sex", "hand"):
            value = _map_options(
                what=col_name, key=value[row_ind], fro="bids", to="mne"
            )
            # We don't know how to translate to MNE, so skip.
            if value is None:
                if col_name == "sex":
                    info_str = "subject sex"
                else:
                    info_str = "subject handedness"
                warn(
                    f'Unable to map "{col_name}" value "{value}" to MNE. '
                    f"Not setting {info_str}."
                )
        elif col_name in ("height", "weight"):
            try:
                value = float(value[row_ind])
            except ValueError:
                value = None
        else:
            if value[row_ind] == "n/a":
                value = None
            else:
                value = value[row_ind]

        # add data into raw.Info
        key = "his_id" if col_name == "participant_id" else col_name
        if value is not None:
            assert key not in raw.info["subject_info"]
            raw.info["subject_info"][key] = value

    return raw


def _handle_scans_reading(scans_fname, raw, bids_path):
    """Read associated scans.tsv and set meas_date."""
    scans_tsv = _from_tsv(scans_fname)
    fname = bids_path.fpath.name

    if fname.endswith(".pdf"):
        # for BTi files, the scan is an entire directory
        fname = fname.split(".")[0]

    # get the row corresponding to the file
    # use string concatenation instead of os.path
    # to work nicely with windows
    data_fname = Path(bids_path.datatype) / fname
    fnames = scans_tsv["filename"]
    fnames = [Path(fname) for fname in fnames]
    if "acq_time" in scans_tsv:
        acq_times = scans_tsv["acq_time"]
    else:
        acq_times = ["n/a"] * len(fnames)

    # There are three possible extensions for BrainVision
    # First gather all the possible extensions
    acq_suffixes = set(fname.suffix for fname in fnames)
    # Add the filename extension for the bids folder
    acq_suffixes.add(Path(data_fname).suffix)

    if all(suffix in (".vhdr", ".eeg", ".vmrk") for suffix in acq_suffixes):
        ext = fnames[0].suffix
        data_fname = Path(data_fname).with_suffix(ext)
    row_ind = _verbose_list_index(fnames, data_fname)

    # check whether all split files have the same acq_time
    # and throw an error if they don't
    if "_split-" in fname:
        split_idx = fname.find("split-")
        pattern = re.compile(
            bids_path.datatype
            + "/"
            + bids_path.basename[:split_idx]
            + r"split-\d+_"
            + bids_path.datatype
            + bids_path.fpath.suffix
        )
        split_fnames = list(filter(lambda x: pattern.match(x.as_posix()), fnames))
        split_acq_times = []
        for split_f in split_fnames:
            split_acq_times.append(acq_times[_verbose_list_index(fnames, split_f)])
        if len(set(split_acq_times)) != 1:
            raise ValueError("Split files must have the same acq_time.")

    # extract the acquisition time from scans file
    acq_time = acq_times[row_ind]
    if acq_time != "n/a":
        # BIDS allows the time to be stored in UTC with a zero time-zone offset, as
        # indicated by a trailing "Z" in the datetime string. If the "Z" is missing, the
        # time is represented as "local" time. We have no way to know what the local
        # time zone is at the *acquisition* site; so we simply assume the same time zone
        # as the user's current system (this is what the spec demands anyway).
        acq_time_is_utc = acq_time.endswith("Z")

        # microseconds part in the acquisition time is optional; add it if missing
        if "." not in acq_time:
            if acq_time_is_utc:
                acq_time = acq_time.replace("Z", ".0Z")
            else:
                acq_time += ".0"

        date_format = "%Y-%m-%dT%H:%M:%S.%f"
        if acq_time_is_utc:
            date_format += "Z"

        acq_time = datetime.strptime(acq_time, date_format)

        if acq_time_is_utc:
            # Enforce setting timezone to UTC without additonal conversion
            acq_time = acq_time.replace(tzinfo=timezone.utc)
        else:
            # Convert time offset to UTC
            acq_time = acq_time.astimezone(timezone.utc)

        logger.debug(
            f"Loaded {scans_fname} scans file to set " f"acq_time as {acq_time}."
        )
        # First set measurement date to None and then call call anonymize() to
        # remove any traces of the measurement date we wish
        # to replace – it might lurk out in more places than just
        # raw.info['meas_date'], e.g. in info['meas_id]['secs'] and in
        # info['file_id'], which are not affected by set_meas_date().
        # The combined use of set_meas_date(None) and anonymize() is suggested
        # by the MNE documentation, and in fact we cannot load e.g. OpenNeuro
        # ds003392 without this combination.
        raw.set_meas_date(None)
        raw.anonymize(daysback=None, keep_his=True)
        raw.set_meas_date(acq_time)

    return raw


def _handle_info_reading(sidecar_fname, raw):
    """Read associated sidecar JSON and populate raw.

    Handle PowerLineFrequency of recording.
    """
    with open(sidecar_fname, encoding="utf-8-sig") as fin:
        sidecar_json = json.load(fin)

    # read in the sidecar JSON's and raw object's line frequency
    json_linefreq = sidecar_json.get("PowerLineFrequency")
    raw_linefreq = raw.info["line_freq"]

    # If both are defined, warn if there is a conflict, else all is fine
    if (json_linefreq is not None) and (raw_linefreq is not None):
        if json_linefreq != raw_linefreq:
            msg = (
                f"Line frequency in sidecar JSON does not match the info "
                f"data structure of the mne.Raw object:\n"
                f"Sidecar JSON is -> {json_linefreq}\n"
                f"Raw is -> {raw_linefreq}\n\n"
            )

            if json_linefreq == "n/a":
                msg += "Defaulting to the info from mne.Raw object."
                raw.info["line_freq"] = raw_linefreq
            else:
                msg += "Defaulting to the info from sidecar JSON."
                raw.info["line_freq"] = json_linefreq

            warn(msg)

    # Else, try to use JSON, fall back on mne.Raw
    elif (json_linefreq is not None) and (json_linefreq != "n/a"):
        raw.info["line_freq"] = json_linefreq
    else:
        pass  # line freq is either defined or None in mne.Raw

    # get cHPI info
    chpi = sidecar_json.get("ContinuousHeadLocalization")
    if chpi is None:
        # no cHPI info in the sidecar – leave raw.info unchanged
        pass
    elif chpi is True:
        from mne.io.ctf import RawCTF
        from mne.io.kit.kit import RawKIT

        msg = (
            "Cannot verify that the cHPI frequencies from "
            "the MEG JSON sidecar file correspond to the raw data{}"
        )

        if isinstance(raw, RawCTF):
            # Pick channels corresponding to the cHPI positions
            hpi_picks = pick_channels_regexp(raw.info["ch_names"], "HLC00[123][123].*")
            if len(hpi_picks) != 9:
                raise ValueError(
                    f"Could not find all cHPI channels that we expected for "
                    f"CTF data. Expected: 9, found: {len(hpi_picks)}"
                )
            logger.info(msg.format(" for CTF files."))

        elif isinstance(raw, RawKIT):
            logger.info(msg.format(" for KIT files."))

        elif "HeadCoilFrequency" in sidecar_json:
            hpi_freqs_json = sidecar_json["HeadCoilFrequency"]
            try:
                hpi_freqs_raw, _, _ = mne.chpi.get_chpi_info(raw.info)
            except ValueError:
                logger.info(msg.format("."))
            else:
                # XXX: Set chpi info in mne.Raw to what is in the sidecar
                if not np.allclose(hpi_freqs_json, hpi_freqs_raw):
                    warn(
                        f"The cHPI coil frequencies in the sidecar file "
                        f"{sidecar_fname}:\n    {hpi_freqs_json}\n "
                        f"differ from what is stored in the raw data:\n"
                        f"    {hpi_freqs_raw}.\n"
                        f"Defaulting to the info from mne.Raw object."
                    )
        else:
            addmsg = (
                ".\n(Because no 'HeadCoilFrequency' data was found in the sidecar.)"
            )
            logger.info(msg.format(addmsg))

    else:
        if raw.info["hpi_subsystem"]:
            logger.info(
                "Dropping cHPI information stored in raw data, "
                "following specification in sidecar file"
            )
        with raw.info._unlock():
            raw.info["hpi_subsystem"] = None
            raw.info["hpi_meas"] = []

    return raw


def _handle_events_reading(events_fname, raw):
    """Read associated events.tsv and populate raw.

    Handle onset, duration, and description of each event.
    """
    logger.info(f"Reading events from {events_fname}.")
    events_dict = _from_tsv(events_fname)

    # Get the descriptions of the events
    if "trial_type" in events_dict:
        trial_type_col_name = "trial_type"
    elif "stim_type" in events_dict:  # Backward-compat with old datasets.
        trial_type_col_name = "stim_type"
        warn(
            f'The events file, {events_fname}, contains a "stim_type" '
            f'column. This column should be renamed to "trial_type" for '
            f"BIDS compatibility."
        )
    else:
        trial_type_col_name = None

    if trial_type_col_name is not None:
        # Drop events unrelated to a trial type
        events_dict = _drop(events_dict, "n/a", trial_type_col_name)

        if "value" in events_dict:
            # Check whether the `trial_type` <> `value` mapping is unique.
            trial_types = events_dict[trial_type_col_name]
            values = np.asarray(events_dict["value"], dtype=str)
            for trial_type in np.unique(trial_types):
                idx = np.where(trial_type == np.atleast_1d(trial_types))[0]
                matching_values = values[idx]

                if len(np.unique(matching_values)) > 1:
                    # Event type descriptors are ambiguous; create hierarchical
                    # event descriptors.
                    logger.info(
                        f'The event "{trial_type}" refers to multiple event '
                        f"values. Creating hierarchical event names."
                    )
                    for ii in idx:
                        value = values[ii]
                        value = "na" if value == "n/a" else value
                        new_name = f"{trial_type}/{value}"
                        logger.info(
                            f"    Renaming event: {trial_type} -> " f"{new_name}"
                        )
                        trial_types[ii] = new_name
            descriptions = np.asarray(trial_types, dtype=str)
        else:
            descriptions = np.asarray(events_dict[trial_type_col_name], dtype=str)
    elif "value" in events_dict:
        # If we don't have a proper description of the events, perhaps we have
        # at least an event value?
        # Drop events unrelated to value
        events_dict = _drop(events_dict, "n/a", "value")
        descriptions = np.asarray(events_dict["value"], dtype=str)

    # Worst case, we go with 'n/a' for all events
    else:
        descriptions = np.array(["n/a"] * len(events_dict["onset"]), dtype=str)

    # Deal with "n/a" strings before converting to float
    onsets = np.array(
        [np.nan if on == "n/a" else on for on in events_dict["onset"]], dtype=float
    )
    durations = np.array(
        [0 if du == "n/a" else du for du in events_dict["duration"]], dtype=float
    )

    # Keep only events where onset is known
    good_events_idx = ~np.isnan(onsets)
    onsets = onsets[good_events_idx]
    durations = durations[good_events_idx]
    descriptions = descriptions[good_events_idx]
    del good_events_idx

    # Add events as Annotations, but keep essential Annotations present in
    # raw file
    annot_from_raw = raw.annotations.copy()

    annot_from_events = mne.Annotations(
        onset=onsets, duration=durations, description=descriptions
    )
    raw.set_annotations(annot_from_events)

    annot_idx_to_keep = [
        idx
        for idx, descr in enumerate(annot_from_raw.description)
        if descr in ANNOTATIONS_TO_KEEP
    ]
    annot_to_keep = annot_from_raw[annot_idx_to_keep]

    if len(annot_to_keep):
        raw.set_annotations(raw.annotations + annot_to_keep)

    return raw


def _get_bads_from_tsv_data(tsv_data):
    """Extract names of bads from data read from channels.tsv."""
    idx = []
    for ch_idx, status in enumerate(tsv_data["status"]):
        if status.lower() == "bad":
            idx.append(ch_idx)

    bads = [tsv_data["name"][i] for i in idx]
    return bads


def _handle_channels_reading(channels_fname, raw):
    """Read associated channels.tsv and populate raw.

    Updates status (bad) and types of channels.
    """
    logger.info(f"Reading channel info from {channels_fname}.")
    channels_dict = _from_tsv(channels_fname)
    ch_names_tsv = channels_dict["name"]

    # Now we can do some work.
    # The "type" column is mandatory in BIDS. We can use it to set channel
    # types in the raw data using a mapping between channel types
    channel_type_bids_mne_map = dict()

    # Get the best mapping we currently have from BIDS to MNE nomenclature
    bids_to_mne_ch_types = _get_ch_type_mapping(fro="bids", to="mne")
    ch_types_json = channels_dict["type"]
    for ch_name, ch_type in zip(ch_names_tsv, ch_types_json):
        # We don't map MEG channels for now, as there's no clear 1:1 mapping
        # from BIDS to MNE coil types.
        if ch_type.upper() in (
            "MEGGRADAXIAL",
            "MEGMAG",
            "MEGREFGRADAXIAL",
            "MEGGRADPLANAR",
            "MEGREFMAG",
            "MEGOTHER",
        ):
            continue

        # Try to map from BIDS nomenclature to MNE, leave channel type
        # untouched if we are uncertain
        updated_ch_type = bids_to_mne_ch_types.get(ch_type, None)

        if updated_ch_type is None:
            # XXX Try again with uppercase spelling – this should be removed
            # XXX once https://github.com/bids-standard/bids-validator/issues/1018  # noqa:E501
            # XXX has been resolved.
            # XXX x-ref https://github.com/mne-tools/mne-bids/issues/481
            updated_ch_type = bids_to_mne_ch_types.get(ch_type.upper(), None)
            if updated_ch_type is not None:
                msg = (
                    "The BIDS dataset contains channel types in lowercase "
                    "spelling. This violates the BIDS specification and "
                    "will raise an error in the future."
                )
                warn(msg)

        if updated_ch_type is None:
            # We don't have an appropriate mapping, so make it a "misc" channel
            channel_type_bids_mne_map[ch_name] = "misc"
            warn(
                f'No BIDS -> MNE mapping found for channel type "{ch_type}". '
                f'Type of channel "{ch_name}" will be set to "misc".'
            )
        else:
            # We found a mapping, so use it
            channel_type_bids_mne_map[ch_name] = updated_ch_type

    # Special handling for (synthesized) stimulus channel
    synthesized_stim_ch_name = "STI 014"
    if (
        synthesized_stim_ch_name in raw.ch_names
        and synthesized_stim_ch_name not in ch_names_tsv
    ):
        logger.info(
            f'The stimulus channel "{synthesized_stim_ch_name}" is present in '
            f"the raw data, but not included in channels.tsv. Removing the "
            f"channel."
        )
        raw.drop_channels([synthesized_stim_ch_name])

    # Rename channels in loaded Raw to match those read from the BIDS sidecar
    if len(ch_names_tsv) != len(raw.ch_names):
        warn(
            f"The number of channels in the channels.tsv sidecar file "
            f"({len(ch_names_tsv)}) does not match the number of channels "
            f"in the raw data file ({len(raw.ch_names)}). Will not try to "
            f"set channel names."
        )
    else:
        raw.rename_channels(dict(zip(raw.ch_names, ch_names_tsv)))

    # Set the channel types in the raw data according to channels.tsv
    channel_type_bids_mne_map_available_channels = {
        ch_name: ch_type
        for ch_name, ch_type in channel_type_bids_mne_map.items()
        if ch_name in raw.ch_names
    }
    ch_diff = set(channel_type_bids_mne_map.keys()) - set(
        channel_type_bids_mne_map_available_channels.keys()
    )
    if ch_diff:
        warn(
            f"Cannot set channel type for the following channels, as they "
            f'are missing in the raw data: {", ".join(sorted(ch_diff))}'
        )
    raw.set_channel_types(channel_type_bids_mne_map_available_channels)

    # Set bad channels based on _channels.tsv sidecar
    if "status" in channels_dict:
        bads_tsv = _get_bads_from_tsv_data(channels_dict)
        bads_avail = [ch_name for ch_name in bads_tsv if ch_name in raw.ch_names]

        ch_diff = set(bads_tsv) - set(bads_avail)
        if ch_diff:
            warn(
                f'Cannot set "bad" status for the following channels, as '
                f"they are missing in the raw data: "
                f'{", ".join(sorted(ch_diff))}'
            )

        raw.info["bads"] = bads_avail

    return raw


@verbose
def read_raw_bids(bids_path, extra_params=None, verbose=None):
    """Read BIDS compatible data.

    Will attempt to read associated events.tsv and channels.tsv files to
    populate the returned raw object with raw.annotations and raw.info['bads'].

    Parameters
    ----------
    bids_path : BIDSPath
        The file to read. The :class:`mne_bids.BIDSPath` instance passed here
        **must** have the ``.root`` attribute set. The ``.datatype`` attribute
        **may** be set. If ``.datatype`` is not set and only one data type
        (e.g., only EEG or MEG data) is present in the dataset, it will be
        selected automatically.

        .. note::
           If ``bids_path`` points to a symbolic link of a ``.fif`` file
           without a ``split`` entity, the link will be resolved before
           reading.

    extra_params : None | dict
        Extra parameters to be passed to MNE read_raw_* functions.
        Note that the ``exclude`` parameter, which is supported by some
        MNE-Python readers, is not supported; instead, you need to subset
        your channels **after** reading.
    %(verbose)s

    Returns
    -------
    raw : mne.io.Raw
        The data as MNE-Python Raw object.

    Raises
    ------
    RuntimeError
        If multiple recording data types are present in the dataset, but
        ``datatype=None``.

    RuntimeError
        If more than one data files exist for the specified recording.

    RuntimeError
        If no data file in a supported format can be located.

    ValueError
        If the specified ``datatype`` cannot be found in the dataset.

    """
    if not isinstance(bids_path, BIDSPath):
        raise RuntimeError(
            '"bids_path" must be a BIDSPath object. Please '
            "instantiate using mne_bids.BIDSPath()."
        )

    bids_path = bids_path.copy()
    sub = bids_path.subject
    ses = bids_path.session
    bids_root = bids_path.root
    datatype = bids_path.datatype
    suffix = bids_path.suffix

    # check root available
    if bids_root is None:
        raise ValueError(
            'The root of the "bids_path" must be set. '
            'Please use `bids_path.update(root="<root>")` '
            "to set the root of the BIDS folder to read."
        )

    # infer the datatype and suffix if they are not present in the BIDSPath
    if datatype is None:
        datatype = _infer_datatype(root=bids_root, sub=sub, ses=ses)
        bids_path.update(datatype=datatype)
    if suffix is None:
        bids_path.update(suffix=datatype)

    if bids_path.fpath.suffix == ".pdf":
        bids_raw_folder = bids_path.directory / f"{bids_path.basename}"

        # try to find the processed data file ("pdf")
        # see: https://www.fieldtriptoolbox.org/getting_started/bti/
        bti_pdf_patterns = ["0", "c,rf*", "hc,rf*", "e,rf*"]
        pdf_list = []
        for pattern in bti_pdf_patterns:
            pdf_list += sorted(bids_raw_folder.glob(pattern))

        if len(pdf_list) == 0:
            raise RuntimeError(
                "Cannot find BTi 'processed data file' (pdf). Please open an issue on "
                "the mne-bids repository to discuss with the developers:\n\n"
                "https://github.com/mne-tools/mne-bids/issues/new/choose\n\n"
                f"No matches for following patterns:\n\n{bti_pdf_patterns}\n\n"
                f"In: {bids_raw_folder}"
            )
        elif len(pdf_list) > 1:  # pragma: no cover
            logger.warn(
                "Found more than one BTi 'processed data file' (pdf). "
                f"Picking:\n\n    {pdf_list[0]}\n\nout of the options:\n\n"
                f"{pdf_list}\n\n"
            )
        raw_path = pdf_list[0]
        config_path = bids_raw_folder / "config"
    else:
        raw_path = bids_path.fpath
        # Resolve for FIFF files
        if (
            raw_path.suffix == ".fif"
            and bids_path.split is None
            and raw_path.is_symlink()
        ):
            target_path = raw_path.resolve()
            logger.info(f"Resolving symbolic link: " f"{raw_path} -> {target_path}")
            raw_path = target_path
        config_path = None

    # Special-handle EDF filenames: we accept upper- and lower-case extensions
    if raw_path.suffix.lower() == ".edf":
        for extension in (".edf", ".EDF"):
            candidate_path = raw_path.with_suffix(extension)
            if candidate_path.exists():
                raw_path = candidate_path
                break

    if not raw_path.exists():
        options = os.listdir(bids_path.directory)
        matches = get_close_matches(bids_path.basename, options)
        msg = f"File does not exist:\n{raw_path}"
        if matches:
            msg += (
                "\nDid you mean one of:\n"
                + "\n".join(matches)
                + "\ninstead of:\n"
                + bids_path.basename
            )
        raise FileNotFoundError(msg)
    if config_path is not None and not config_path.exists():
        raise FileNotFoundError(f"config directory not found: {config_path}")

    if extra_params is None:
        extra_params = dict()
    elif "exclude" in extra_params:
        del extra_params["exclude"]
        logger.info('"exclude" parameter is not supported by read_raw_bids')

    if raw_path.suffix == ".fif" and "allow_maxshield" not in extra_params:
        extra_params["allow_maxshield"] = True
    raw = _read_raw(
        raw_path,
        electrode=None,
        hsp=None,
        hpi=None,
        config_path=config_path,
        **extra_params,
    )

    # Try to find an associated events.tsv to get information about the
    # events in the recorded data
    if (
        bids_path.subject == "emptyroom" and bids_path.task == "noise"
    ) or bids_path.task.startswith("rest"):
        on_error = "ignore"
    else:
        on_error = "warn"

    events_fname = _find_matching_sidecar(
        bids_path, suffix="events", extension=".tsv", on_error=on_error
    )

    if events_fname is not None:
        raw = _handle_events_reading(events_fname, raw)

    # Try to find an associated channels.tsv to get information about the
    # status and type of present channels
    channels_fname = _find_matching_sidecar(
        bids_path, suffix="channels", extension=".tsv", on_error="warn"
    )
    if channels_fname is not None:
        raw = _handle_channels_reading(channels_fname, raw)

    # Try to find an associated electrodes.tsv and coordsystem.json
    # to get information about the status and type of present channels
    on_error = "warn" if suffix == "ieeg" else "ignore"
    electrodes_fname = _find_matching_sidecar(
        bids_path, suffix="electrodes", extension=".tsv", on_error=on_error
    )
    coordsystem_fname = _find_matching_sidecar(
        bids_path, suffix="coordsystem", extension=".json", on_error=on_error
    )
    if electrodes_fname is not None:
        if coordsystem_fname is None:
            raise RuntimeError(
                f"BIDS mandates that the coordsystem.json "
                f"should exist if electrodes.tsv does. "
                f"Please create coordsystem.json for"
                f"{bids_path.basename}"
            )
        if datatype in ["meg", "eeg", "ieeg"]:
            _read_dig_bids(
                electrodes_fname, coordsystem_fname, raw=raw, datatype=datatype
            )

    # Try to find an associated sidecar .json to get information about the
    # recording snapshot
    sidecar_fname = _find_matching_sidecar(
        bids_path, suffix=datatype, extension=".json", on_error="warn"
    )
    if sidecar_fname is not None:
        raw = _handle_info_reading(sidecar_fname, raw)

    # read in associated scans filename
    scans_fname = BIDSPath(
        subject=bids_path.subject,
        session=bids_path.session,
        suffix="scans",
        extension=".tsv",
        root=bids_path.root,
    ).fpath

    if scans_fname.exists():
        raw = _handle_scans_reading(scans_fname, raw, bids_path)

    # read in associated subject info from participants.tsv
    participants_tsv_path = bids_root / "participants.tsv"
    subject = f"sub-{bids_path.subject}"
    if op.exists(participants_tsv_path):
        raw = _handle_participants_reading(
            participants_fname=participants_tsv_path, raw=raw, subject=subject
        )
    else:
        warn(f"participants.tsv file not found for {raw_path}")
        raw.info["subject_info"] = dict()

    assert raw.annotations.orig_time == raw.info["meas_date"]
    return raw


@verbose
def get_head_mri_trans(
    bids_path,
    extra_params=None,
    t1_bids_path=None,
    fs_subject=None,
    fs_subjects_dir=None,
    *,
    kind=None,
    verbose=None,
):
    """Produce transformation matrix from MEG and MRI landmark points.

    Will attempt to read the landmarks of Nasion, LPA, and RPA from the sidecar
    files of (i) the MEG and (ii) the T1-weighted MRI data. The two sets of
    points will then be used to calculate a transformation matrix from head
    coordinates to MRI coordinates.

    .. note:: The MEG and MRI data need **not** necessarily be stored in the
              same session or even in the same BIDS dataset. See the
              ``t1_bids_path`` parameter for details.

    Parameters
    ----------
    bids_path : BIDSPath
        The path of the electrophysiology recording. If ``datatype`` and
        ``suffix`` are not present, they will be set to ``'meg'``, and a
        warning will be raised.

        .. versionchanged:: 0.10
           A warning is raised it ``datatype`` or ``suffix`` are not set.
    extra_params : None | dict
        Extra parameters to be passed to :func:`mne.io.read_raw` when reading
        the MEG file.
    t1_bids_path : BIDSPath | None
        If ``None`` (default), will try to discover the T1-weighted MRI file
        based on the name and location of the MEG recording specified via the
        ``bids_path`` parameter. Alternatively, you explicitly specify which
        T1-weighted MRI scan to use for extraction of MRI landmarks. To do
        that, pass a :class:`mne_bids.BIDSPath` pointing to the scan.
        Use this parameter e.g. if the T1 scan was recorded during a different
        session than the MEG. It is even possible to point to a T1 image stored
        in an entirely different BIDS dataset than the MEG data.
    fs_subject : str
        The subject identifier used for FreeSurfer.

        .. versionchanged:: 0.10
           Does not default anymore to ``bids_path.subject`` if ``None``.
    fs_subjects_dir : path-like | None
        The FreeSurfer subjects directory. If ``None``, defaults to the
        ``SUBJECTS_DIR`` environment variable.

        .. versionadded:: 0.8
    kind : str | None
        The suffix of the anatomical landmark names in the JSON sidecar.
        A suffix might be present e.g. to distinguish landmarks between
        sessions. If provided, should not include a leading underscore ``_``.
        For example, if the landmark names in the JSON sidecar file are
        ``LPA_ses-1``, ``RPA_ses-1``, ``NAS_ses-1``, you should pass
        ``'ses-1'`` here.
        If ``None``, no suffix is appended, the landmarks named
        ``Nasion`` (or ``NAS``), ``LPA``, and ``RPA`` will be used.

        .. versionadded:: 0.10
    %(verbose)s

    Returns
    -------
    trans : mne.transforms.Transform
        The data transformation matrix from head to MRI coordinates.
    """
    nib = _import_nibabel("get a head to MRI transform")

    if not isinstance(bids_path, BIDSPath):
        raise RuntimeError(
            '"bids_path" must be a BIDSPath object. Please '
            "instantiate using mne_bids.BIDSPath()."
        )

    # check root available
    meg_bids_path = bids_path.copy()
    del bids_path
    if meg_bids_path.root is None:
        raise ValueError(
            'The root of the "bids_path" must be set. '
            'Please use `bids_path.update(root="<root>")` '
            "to set the root of the BIDS folder to read."
        )

    # if the bids_path is underspecified, only get info for MEG data
    if meg_bids_path.datatype is None:
        meg_bids_path.datatype = "meg"
        warn(
            'bids_path did not have a datatype set. Assuming "meg". This '
            "will raise an exception in the future.",
            module="mne_bids",
            category=DeprecationWarning,
        )
    if meg_bids_path.suffix is None:
        meg_bids_path.suffix = "meg"
        warn(
            'bids_path did not have a suffix set. Assuming "meg". This '
            "will raise an exception in the future.",
            module="mne_bids",
            category=DeprecationWarning,
        )

    # Get the sidecar file for MRI landmarks
    t1w_bids_path = (
        (meg_bids_path if t1_bids_path is None else t1_bids_path)
        .copy()
        .update(datatype="anat", suffix="T1w", task=None)
    )
    t1w_json_path = _find_matching_sidecar(
        bids_path=t1w_bids_path, extension=".json", on_error="ignore"
    )
    del t1_bids_path

    if t1w_json_path is not None:
        t1w_json_path = Path(t1w_json_path)

    if t1w_json_path is None or not t1w_json_path.exists():
        raise FileNotFoundError(
            f"Did not find T1w JSON sidecar file, tried location: " f"{t1w_json_path}"
        )
    for extension in (".nii", ".nii.gz"):
        t1w_path_candidate = t1w_json_path.with_suffix(extension)
        if t1w_path_candidate.exists():
            t1w_bids_path = get_bids_path_from_fname(fname=t1w_path_candidate)
            break

    if not t1w_bids_path.fpath.exists():
        raise FileNotFoundError(
            f"Did not find T1w recording file, tried location: "
            f'{t1w_path_candidate.name.replace(".nii.gz", "")}[.nii, .nii.gz]'
        )

    # Get MRI landmarks from the JSON sidecar
    t1w_json = json.loads(t1w_json_path.read_text(encoding="utf-8"))
    mri_coords_dict = t1w_json.get("AnatomicalLandmarkCoordinates", dict())

    # landmarks array: rows: [LPA, NAS, RPA]; columns: [x, y, z]
    suffix = f"_{kind}" if kind is not None else ""
    mri_landmarks = np.full((3, 3), np.nan)
    for landmark_name, coords in mri_coords_dict.items():
        if landmark_name.upper() == ("LPA" + suffix).upper():
            mri_landmarks[0, :] = coords
        elif landmark_name.upper() == ("RPA" + suffix).upper():
            mri_landmarks[2, :] = coords
        elif (
            landmark_name.upper() == ("NAS" + suffix).upper()
            or landmark_name.lower() == ("nasion" + suffix).lower()
        ):
            mri_landmarks[1, :] = coords
        else:
            continue

    if np.isnan(mri_landmarks).any():
        raise RuntimeError(
            f"Could not extract fiducial points from T1w sidecar file: "
            f"{t1w_json_path}\n\n"
            f"The sidecar file SHOULD contain a key "
            f'"AnatomicalLandmarkCoordinates" pointing to an '
            f'object with the keys "LPA", "NAS", and "RPA". '
            f"Yet, the following structure was found:\n\n"
            f"{mri_coords_dict}"
        )

    # The MRI landmarks are in "voxels". We need to convert them to the
    # Neuromag RAS coordinate system in order to compare them with MEG
    # landmarks. See also: `mne_bids.write.write_anat`
    if fs_subject is None:
        warn(
            'Passing "fs_subject=None" has been deprecated and will raise '
            "an error in future versions. Please explicitly specify the "
            "FreeSurfer subject name.",
            DeprecationWarning,
        )
        fs_subject = f"sub-{meg_bids_path.subject}"

    fs_subjects_dir = get_subjects_dir(fs_subjects_dir, raise_error=False)
    fs_t1_path = Path(fs_subjects_dir) / fs_subject / "mri" / "T1.mgz"
    if not fs_t1_path.exists():
        raise ValueError(
            f"Could not find {fs_t1_path}. Consider running FreeSurfer's "
            f"'recon-all` for subject {fs_subject}."
        )
    fs_t1_mgh = nib.load(str(fs_t1_path))
    t1_nifti = nib.load(str(t1w_bids_path.fpath))

    # Convert to MGH format to access vox2ras method
    t1_mgh = nib.MGHImage(t1_nifti.dataobj, t1_nifti.affine)

    # convert to scanner RAS
    mri_landmarks = apply_trans(t1_mgh.header.get_vox2ras(), mri_landmarks)

    # convert to FreeSurfer T1 voxels (same scanner RAS as T1)
    mri_landmarks = apply_trans(fs_t1_mgh.header.get_ras2vox(), mri_landmarks)

    # now extract transformation matrix and put back to RAS coordinates of MRI
    vox2ras_tkr = fs_t1_mgh.header.get_vox2ras_tkr()
    mri_landmarks = apply_trans(vox2ras_tkr, mri_landmarks)
    mri_landmarks = mri_landmarks * 1e-3

    # Get MEG landmarks from the raw file
    _, ext = _parse_ext(meg_bids_path)
    if extra_params is None:
        extra_params = dict()
    if ext == ".fif":
        extra_params["allow_maxshield"] = "yes"

    raw = read_raw_bids(bids_path=meg_bids_path, extra_params=extra_params)

    if (
        raw.get_montage() is None
        or raw.get_montage().get_positions() is None
        or any(
            [
                raw.get_montage().get_positions()[fid_key] is None
                for fid_key in ("nasion", "lpa", "rpa")
            ]
        )
    ):
        raise RuntimeError(
            f"Could not extract fiducial points from ``raw`` file: "
            f"{meg_bids_path}\n\n"
            f"The ``raw`` file SHOULD contain digitization points "
            "for the nasion and left and right pre-auricular points "
            "but none were found"
        )
    pos = raw.get_montage().get_positions()
    meg_landmarks = np.asarray((pos["lpa"], pos["nasion"], pos["rpa"]))

    # Given the two sets of points, fit the transform
    trans_fitted = fit_matched_points(src_pts=meg_landmarks, tgt_pts=mri_landmarks)
    trans = mne.transforms.Transform(fro="head", to="mri", trans=trans_fitted)
    return trans



"""Count events in BIDS dataset.

example usage:  $ mne_bids count_events --bids_root bids_root_path

"""

# Authors: Alex Gramfort <alexandre.gramfort@inria.fr>
#
# License: BSD-3-Clause
from pathlib import Path

import mne_bids
from mne_bids.stats import count_events


def run():
    """Run the raw_to_bids command."""
    import pandas as pd
    from mne.commands.utils import get_optparser

    parser = get_optparser(
        __file__,
        usage="usage: %prog options args",
        prog_prefix="mne_bids",
        version=mne_bids.__version__,
    )

    parser.add_option(
        "--bids_root", dest="bids_root", help="The path of the BIDS compatible folder."
    )

    parser.add_option(
        "--datatype", dest="datatype", default="auto", help="The datatype to consider."
    )

    parser.add_option(
        "--describe",
        dest="describe",
        action="store_true",
        help="If set print the descriptive statistics (min, max, etc.).",
    )

    parser.add_option(
        "--output",
        dest="output",
        default=None,
        help="Path to the CSV file where to store the results.",
    )

    parser.add_option(
        "--overwrite",
        dest="overwrite",
        action="store_true",
        help="If set, overwrite an existing output file.",
    )

    parser.add_option(
        "--silent",
        dest="silent",
        action="store_true",
        help="Whether to print the event counts on the screen.",
    )

    opt, args = parser.parse_args()

    if len(args) > 0:
        parser.print_help()
        parser.error(f'Do not specify arguments without flags. Found: "{args}".\n')

    if not all([opt.bids_root]):
        parser.print_help()
        parser.error(
            "Arguments missing. You need to specify the --bids_root parameter."
        )

    if opt.output and Path(opt.output).exists() and not opt.overwrite:
        parser.error("Output file exists. To overwrite, pass --overwrite")

    counts = count_events(opt.bids_root, datatype=opt.datatype)

    if opt.describe:
        counts = counts.describe()

    if not opt.silent:
        with pd.option_context(
            "display.max_rows", 1000, "display.max_columns", 80, "display.width", 1000
        ):
            print(counts)

    if opt.output:
        counts.to_csv(opt.output)
        print(f"\nOutput stored in {opt.output}")


if __name__ == "__main__":
    run()



"""MNE software for easily interacting with BIDS compatible datasets."""

try:
    from importlib.metadata import version

    __version__ = version("mne_bids")
except Exception:
    __version__ = "0.0.0"

from mne_bids import commands
from mne_bids.report import make_report
from mne_bids.path import (
    BIDSPath,
    get_datatypes,
    get_entity_vals,
    print_dir_tree,
    get_entities_from_fname,
    search_folder_for_text,
    get_bids_path_from_fname,
    find_matching_paths,
)
from mne_bids.read import get_head_mri_trans, read_raw_bids
from mne_bids.utils import get_anonymization_daysback
from mne_bids.write import (
    make_dataset_description,
    write_anat,
    write_raw_bids,
    mark_channels,
    write_meg_calibration,
    write_meg_crosstalk,
    get_anat_landmarks,
    anonymize_dataset,
)
from mne_bids.sidecar_updates import update_sidecar_json, update_anat_landmarks
from mne_bids.inspect import inspect_dataset
from mne_bids.dig import (
    template_to_head,
    convert_montage_to_ras,
    convert_montage_to_mri,
)
